from __future__ import absolute_import

import os
import json
import codecs
import shutil
import tempfile
import contextlib
import subprocess   # nosec - see usage below
import collections

import boto3
from botocore.client import Config

import requests
from requests_toolbelt.multipart.encoder import MultipartEncoder

from celery import shared_task
from celery.exceptions import MaxRetriesExceededError
from celery.utils.log import get_task_logger

from django.conf import settings
from django.contrib.staticfiles import finders
from django.template import loader
from django.utils.crypto import get_random_string

from regulations.models import FailedCommentSubmission
logger = get_task_logger(__name__)


@shared_task(bind=True, acks_late=True)
def submit_comment(self, comments, form_data, metadata_url):
    '''
    Submit the comment to regulations.gov. If unsuccessful, retry the task.
    Number of retries and time between retries is managed by Celery settings.
    The main comment is converted to a PDF and added as an attachment; the
    'general_comment' field refers to this attachment.

    :param comments: List of sectional comments
    :param form_data: Dict of fields accepted by regulations.gov
    :param metadata_url: SignedUrl for comment metadata

    :return: On success: {"trackingNumber": "...", "pdfUrl": "..."}
             On failure: raises "retry" exception
             On multiple failures: {"error": "Message"}
    '''
    try:
        html = json_to_html(comments)
        files = extract_files(comments)
        with html_to_pdf(html) as comment_pdf, \
                build_attachments(files) as attachments:
            document_number = get_document_number(comments)
            pdf_url = cache_pdf(comment_pdf, document_number, metadata_url)

            # Restore file position changed by cache_pdf
            comment_pdf.seek(0)

            data = build_multipart_encoded(
                form_data, comment_pdf, attachments)
            response = post_submission(data)
            if response.status_code != requests.codes.created:
                logger.warn("Post to regulations.gov failed: %s %s",
                            response.status_code, response.text)
                raise self.retry()
            logger.info(response.text)
            data = response.json()
            return {
                'trackingNumber': data['trackingNumber'],
                'pdfUrl': pdf_url.url,
            }
    except Exception:   # We want to catch _any_ exception
        logger.exception("Unable to submit comment")
        try:
            raise self.retry()
        except MaxRetriesExceededError:
            message = "Exceeded retries, saving failed submission"
            logger.error(message)
            save_failed_submission(
                json.dumps({'comments': comments, 'form_data': form_data})
            )
            return {'error': message}


@shared_task(acks_late=True)
def publish_tracking_number(response, metadata_url):
    """Write the tracking number to S3. Not ideal if this fails, but the
    comment will have been taken care of already, so no need for additional
    safety checks"""
    s3_client.put_object(
        Body=json.dumps(response).encode(),
        Bucket=settings.ATTACHMENT_BUCKET,
        ContentType='application/json',
        Key=metadata_url.key,
    )


def json_to_html(sections, mark_as_draft=False):
    """
    Render the comment as HTML

    :param sections: The section-by-section comments
    :param mark_as_draft: Indicates if the HTML is to be marked as DRAFT
    """
    return loader.render_to_string(
        'regulations/comment.html',
        {'sections': sections, 'mark_as_draft': mark_as_draft}
    )


@contextlib.contextmanager
def html_to_pdf(html):
    try:
        path = tempfile.mkdtemp()
        html_path = os.path.join(path, 'document.html')
        pdf_path = os.path.join(path, 'document.pdf')
        with codecs.open(html_path, 'w', 'utf-8') as fp:
            fp.write(html)
        # Safe because: user provides no input
        subprocess.check_output([   # nosec
            settings.WKHTMLTOPDF_PATH,
            '--user-style-sheet',
            finders.find('regulations/css/style.css'),
            html_path,
            pdf_path,
        ])
        with open(pdf_path, 'rb') as pdf_file:
            yield pdf_file
    finally:
        shutil.rmtree(path)


def cache_pdf(pdf, document_number, metadata_url):
    """Update submission metadata and cache comment PDF."""
    url = SignedUrl.generate()
    content_disposition = generate_content_disposition(document_number,
                                                       draft=False)
    s3_client.put_object(
        Body=json.dumps({'pdfUrl': metadata_url.url}),
        Bucket=settings.ATTACHMENT_BUCKET,
        Key=metadata_url.key,
    )
    s3_client.put_object(
        Body=pdf,
        ContentType='application/pdf',
        ContentDisposition=content_disposition,
        Bucket=settings.ATTACHMENT_BUCKET,
        Key=url.key,
    )
    return url


def generate_content_disposition(document_number, draft=False):
    return 'attachment; filename="{}comment_{}.pdf"'.format(
        "DRAFT_" if draft else "",
        document_number
    )


def get_document_number(comments):
    """ Get the FR document number for the notice against which the comments
    are being submitted
    """
    return comments[0]['docId']


@contextlib.contextmanager
def build_attachments(files):
    '''
    Assembles a collection of tuples of the form:
    [
        ('uploadedFile', ('fileName1', file-object1),
        ('uploadedFile', ('fileName2', file-object2),
        ...
    ]
    for POSTing as a multipart/form-data upload.
    On context exit, the file objects are closed and the locally
    downloaded files are deleted.
    '''
    attachments = []
    try:
        path = tempfile.mkdtemp()
        attachments = [
            ('uploadedFile', (file_['name'],
                              fetch_file(path, file_['key'], file_['name'])))
            for file_ in files
        ]
        yield attachments
    finally:
        for file_ in attachments:
            file_[1][1].close()
        shutil.rmtree(path)


def fetch_file(path, key, name):
    '''
    Returns a file object corresponding to a local file stored at ``path+name``
    whose content is downloaded from S3 where it is stored under ``key``

    '''
    dest = os.path.join(path, name)
    s3_client.download_file(settings.ATTACHMENT_BUCKET, key, dest)
    return open(dest, "rb")


def make_s3_client():
    session = boto3.Session(
        aws_access_key_id=settings.ATTACHMENT_ACCESS_KEY_ID,
        aws_secret_access_key=settings.ATTACHMENT_SECRET_ACCESS_KEY,
    )
    return session.client('s3', config=Config(signature_version='s3v4'))


s3_client = make_s3_client()


class SignedUrl(collections.namedtuple('SignedUrl', ['key', 'url'])):
    @classmethod
    def generate(cls, key=None, method='get_object', params=None):
        key = key or get_random_string(50)
        params = params or {}
        params.update({'Bucket': settings.ATTACHMENT_BUCKET, 'Key': key})
        url = s3_client.generate_presigned_url(
            ClientMethod=method,
            Params=params,
        )
        return cls(key, url)


def extract_files(sections):
    '''
    Extracts the files that are to be attached to the comment.
    Returns a collection of dicts where for each dict:
        - dict['key'] specifies the file to be attached from S3
        - dict['name'] specifies the name under which the file is to be
          attached.
    '''
    return [
        file
        for section in sections
        for file in section.get('files', [])
    ]


def build_multipart_encoded(form_data, comment_pdf, attachments):
    """ Build a MultiPartEncoded payload from the extra body fields,
        the main comment PDF and the set of attachments
    """
    fields = [
        ('comment_on', settings.COMMENT_DOCUMENT_ID),
        # TODO: Ensure this name is unique
        ('uploadedFile', ('comment.pdf', comment_pdf)),
        ('general_comment', 'See attached comment.pdf'),
    ]

    # Add other submitted fields
    fields.extend(form_data.items())
    fields.extend(attachments)
    return MultipartEncoder(fields)


def save_failed_submission(body):
    FailedCommentSubmission.objects.create(body=body)


def post_submission(data):
    return requests.post(
        settings.REGS_GOV_API_URL,
        data=data,
        headers={
            'Content-Type': data.content_type,
            'X-Api-Key': settings.REGS_GOV_API_KEY,
        }
    )



import re

from six.moves.queue import PriorityQueue
from six.moves.html_parser import HTMLParser

from regulations.generator.layers.location_replace import LocationReplace


class LayersApplier(object):
    """ Most layers replace content. We try to do this intelligently here,
    so that layers don't step over each other. """
    HTML_TAG_REGEX = re.compile(r'<[^>]*?>')

    def __init__(self):
        self.queue = PriorityQueue()
        self.text = None

    def enqueue_from_list(self, elements_list):
        for le in elements_list:
            self.enqueue(le)

    def enqueue(self, layer_element):
        original, replacement, locations = layer_element
        priority = len(original)
        item = (original, replacement, locations)
        self.queue.put((-priority, item))

    def location_replace(self, xml_node, original, replacement, locations):
        LocationReplace().location_replace(xml_node, original, replacement,
                                           locations)

    def unescape_text(self):
        """ Because of the way we do replace_all(), we need to unescape HTML
        entities.  """
        self.text = HTMLParser().unescape(self.text)

    def replace_all(self, original, replacement):
        """ Replace all occurrences of original with replacement. This is HTML
        aware; it effectively looks at all of the text in between HTML tags"""
        text_chunks = []
        index = 0
        for match in self.HTML_TAG_REGEX.finditer(self.text):
            text = self.text[index:match.start()]
            text_chunks.append(text.replace(original, replacement))
            text_chunks.append(self.text[match.start():match.end()])    # tag
            index = match.end()
        text_chunks.append(self.text[index:])   # trailing text
        self.text = "".join(text_chunks)
        self.unescape_text()

    def replace_at(self, original, replacement, locations):
        """ Replace the occurrences of original at all the locations with
        replacement. """

        locations.sort()
        self.text = LocationReplace().location_replace_text(
            self.text, original, replacement, locations)
        self.unescape_text()

    def apply_layers(self, original_text):
        self.text = original_text

        while not self.queue.empty():
            priority, layer_element = self.queue.get()
            original, replacement, locations = layer_element

            if not locations:
                self.replace_all(original, replacement)
            else:
                self.replace_at(original, replacement, locations)

        return self.text


class LayersBase(object):
    """ Base class which keeps track of multiple laeyrs. """
    def __init__(self):
        self.layers = {}

    def add_layer(self, layer):
        self.layers[layer.__class__.shorthand] = layer


class SearchReplaceLayersApplier(LayersBase):
    def get_layer_pairs(self, text_index):
        elements = []
        for layer in self.layers.values():
            elements += list(layer.apply_layer(text_index))
        return elements


class InlineLayersApplier(LayersBase):
    """ Apply multiple inline layers to given text (e.g. links,
    highlighting, etc.) """
    def get_layer_pairs(self, text_index, original_text):
        layer_pairs = []
        for layer in self.layers.values():
            layer_pairs += list(layer.apply_layer(original_text, text_index))

        # convert from offset-based to a search and replace layer.
        layer_elements = []

        for o, r, offset in layer_pairs:
            offset_locations = LocationReplace.find_all_offsets(o,
                                                                original_text)
            locations = [offset_locations.index(offset)]
            layer_elements.append((o, r, locations))
        return layer_elements


class ParagraphLayersApplier(LayersBase):
    """ Handle layers which apply to the whole paragraph. Layers include
    interpretations, section-by-section analyses, table of contents, etc."""

    def apply_layers(self, node):
        for layer in self.layers.values():
            pair = layer.apply_layer(node['markup_id'])
            if pair:
                node[pair[0]] = pair[1]
        return node



from unittest import TestCase
from regulations.generator.layers import tree_builder
from regulations.generator.node_types import REGTEXT

import itertools


class TreeBuilderTest(TestCase):

    def build_tree(self):
        child = {
            'text': 'child text',
            'children': [],
            'label_id': '204-3',
            'label': ['204', '3'],
            'node_type': REGTEXT
        }
        tree = {
            'text': 'parent text',
            'children': [child],
            'label_id': '204',
            'label': ['204'],
            'node_type': REGTEXT
        }
        return tree

    def test_build_tree_hash(self):
        tree = self.build_tree()
        tree['children'][0]['children'] = [{
            'text': 'child child test',
            'children': [],
            'label_id': '204-3-a',
            'label': ['204', '3', 'a'],
            'node_type': REGTEXT
        }]
        tree_hash = tree_builder.build_tree_hash(tree)
        self.assertEqual(set(tree_hash.keys()),
                         set(['204-3-a', '204-3', '204']))

    def test_parent_in_tree(self):
        tree = self.build_tree()
        tree_hash = tree_builder.build_tree_hash(tree)
        self.assertEqual(set(tree_hash.keys()), set(['204-3', '204']))

        self.assertTrue(tree_builder.parent_in_tree('204-3', tree_hash))

    def test_add_node(self):
        new_node = {
            'text': 'new node text',
            'children': [],
            'label_id': '204-4',
            'label': ['204', '4'],
            'node_type': REGTEXT
        }
        tree = self.build_tree()
        tree_hash = tree_builder.build_tree_hash(tree)
        self.assertEqual(set(tree_hash.keys()), set(['204-3', '204']))

        self.assertEqual(len(tree_hash['204']['children']), 1)
        tree_builder.add_node_to_tree(new_node, '204', tree_hash)
        self.assertEqual(len(tree_hash['204']['children']), 2)

        child_labels = [c['label_id'] for c in tree_hash['204']['children']]
        self.assertEqual(child_labels, ['204-3', '204-4'])

    def test_make_label_sortable_roman(self):
        label = "iv"
        sortable = tree_builder.make_label_sortable(label, roman=True)
        self.assertEquals(sortable, (4,))

    def test_make_label_sortable_not_roman(self):
        label = "iv"
        sortable = tree_builder.make_label_sortable(label)
        self.assertEquals(sortable, ('iv',))

    def test_parent_label(self):
        node = {'node_type': 'REGTEXT', 'label': ['204', 'a', '1', 'ii']}
        parent_label = tree_builder.parent_label(node)
        self.assertEquals(['204', 'a', '1'], parent_label)

        node = {'node_type': 'INTERP', 'label': ['204', 'Interp']}
        parent_label = tree_builder.parent_label(node)
        self.assertEquals(['204'], parent_label)

        node = {'node_type': 'INTERP', 'label': ['204', '2', 'Interp']}
        parent_label = tree_builder.parent_label(node)
        self.assertEquals(['204', 'Interp'], parent_label)

        node = {'node_type': 'INTERP', 'label': ['204', '2', 'a', 'Interp']}
        parent_label = tree_builder.parent_label(node)
        self.assertEquals(['204', '2', 'Interp'], parent_label)

        node = {'node_type': 'INTERP',
                'label': ['204', '2', 'Interp', '1']}
        parent_label = tree_builder.parent_label(node)
        self.assertEquals(['204', '2', 'Interp'], parent_label)

        node = {'node_type': 'INTERP',
                'label': ['204', '2', 'Interp', '1', 'i']}
        parent_label = tree_builder.parent_label(node)
        self.assertEquals(['204', '2', 'Interp', '1'], parent_label)

    def test_roman_nums(self):
        first_five = list(itertools.islice(tree_builder.roman_nums(), 0, 5))
        self.assertEquals(['i', 'ii', 'iii', 'iv', 'v'], first_five)

    def test_add_child(self):
        tree = self.build_tree()

        child = {
            'children': [],
            'label': ['204', '2'],
            'label_id': '204-2',
            'node_type': REGTEXT,
            'sortable': (2,),
            'text': 'child text',
        }

        static_child = {
            'children': [],
            'label': ['204', '3'],
            'label_id': '204-3',
            'node_type': REGTEXT,
            'sortable': (3,),
            'text': 'child text',
        }

        static_tree = {
            'children': [child, static_child],
            'text': 'parent text',
            'label': ['204'],
            'label_id': '204',
            'node_type': REGTEXT
        }

        tree_builder.add_child(tree, child)
        self.assertEquals(static_tree, tree)

    def test_add_child_appendix(self):
        parent = {'children': [
            {'node_type': 'APPENDIX', 'label': ['204', 'A', '1']},
            {'node_type': 'APPENDIX', 'label': ['204', 'A', '3']},
            ]}

        child_to_add = {'node_type': 'APPENDIX', 'label': ['204', 'A', '2(a)']}
        tree_builder.add_child(parent, child_to_add)
        self.assertEquals(
            ['204-A-1', '204-A-2(a)', '204-A-3'],
            ['-'.join(c['label']) for c in parent['children']]
            )

    def test_add_child_interp(self):
        parent = {'children': [
            {'node_type': 'INTERP', 'label': ['204', '4', 'Interp']},
            {'node_type': 'INTERP', 'label': ['204', '2', 'Interp']}
        ], 'label': ['204', 'Interp']}
        tree_builder.add_child(parent, {'node_type': 'INTERP',
                                        'label': ['204', '3', 'Interp']})
        self.assertEqual([(2, 2), (2, 3), (2, 4)],
                         [c['sortable'] for c in parent['children']])

        prefix = ['204', '4', 'a', '2']
        parent = {'children': [
            {'node_type': 'INTERP', 'label': prefix + ['v', 'Interp']},
            {'node_type': 'INTERP', 'label': prefix + ['iv', 'Interp']}
        ], 'label': ['204', '4', 'a', '2', 'Interp']}
        tree_builder.add_child(parent, {'node_type': 'INTERP',
                                        'label': prefix + ['ix', 'Interp']})
        self.assertEqual([(4,), (5,), (9,)],
                         [c['sortable'] for c in parent['children']])

        prefix = ['204', '4', 'Interp']
        parent = {'children': [
            {'node_type': 'INTERP', 'label': prefix + ['1']},
            {'node_type': 'INTERP', 'label': prefix + ['3']}
        ], 'label': ['204', 'Interp']}
        tree_builder.add_child(parent, {'node_type': 'INTERP',
                                        'label': prefix + ['2']})
        self.assertEqual([(2, 1), (2, 2), (2, 3)],
                         [c['sortable'] for c in parent['children']])

        prefix = ['204', 'Interp', '2']
        parent = {'children': [
            {'node_type': 'INTERP', 'label': prefix + ['v']},
            {'node_type': 'INTERP', 'label': prefix + ['iv']}
        ], 'label': prefix}
        tree_builder.add_child(parent, {'node_type': 'INTERP',
                                        'label': prefix + ['ix']})
        self.assertEqual([(4,), (5,), (9,)],
                         [c['sortable'] for c in parent['children']])

    def test_add_child_root_interp(self):
        """ Let's add an introductory paragraph child to a root interpretation
        node and ensure that the children are sorted correctly. """

        parent = {'children': [
            {'node_type': 'INTERP', 'label': ['204', '4', 'Interp']},
            {'node_type': 'INTERP', 'label': ['204', '2', 'Interp']}
        ], 'label': ['204', 'Interp']}

        tree_builder.add_child(parent, {'node_type': 'INTERP',
                                        'label': ['204', 'Interp', 'h1']})

        self.assertEqual([(1, 'h', 1), (2, 2), (2, 4)],
                         [c['sortable'] for c in parent['children']])

    def test_add_child_odd_sort(self):
        """Appendices may have some strange orderings. Make sure they keep
        order."""
        parent = {'children': [], 'label': ['204', 'A'],
                  'node_type': 'appendix',
                  'child_labels': ['204-A-p1', '204-A-X', '204-A-L',
                                   '204-A-h1']}

        def mknode(label):
            return {'node_type': 'appendix', 'label': label.split('-')}
        tree_builder.add_child(parent, mknode('204-A-X'))
        tree_builder.add_child(parent, mknode('204-A-L'))
        tree_builder.add_child(parent, mknode('204-A-h1'))
        tree_builder.add_child(parent, mknode('204-A-p1'))
        self.assertEqual([['204', 'A', 'p1'], ['204', 'A', 'X'],
                          ['204', 'A', 'L'], ['204', 'A', 'h1']],
                         [c['label'] for c in parent['children']])

    def test_all_children_are_roman(self):
        parent_node = {'children': [
            {'label': ['201', '4', 'i']},
            {'label': ['201', '4', 'ii']},
            {'label': ['201', '4', 'iii']},
            {'label': ['201', '4', 'iv']},
            {'label': ['201', '4', 'v']},
        ]}

        self.assertTrue(tree_builder.all_children_are_roman(parent_node))

        parent_node = {'children': [
            {'label': ['201', '4', 'a']},
            {'label': ['201', '4', 'i']},
            {'label': ['201', '4', 'v']},
        ]}

        self.assertFalse(tree_builder.all_children_are_roman(parent_node))

    def test_add_child_root_appendix(self):
        """ Let's add an introductory paragraph child to a root interpretation
        node and ensure that the children are sorted correctly. """

        parent = {'children': [
            {'node_type': 'appendix', 'label': ['204', 'A', '4', 'b', 'i']},
            {'node_type': 'appendix', 'label': ['204', 'A', '4', 'b', 'v']}
            ], 'label': ['204', 'appendix']}

        tree_builder.add_child(parent, {'node_type': 'appendix',
                                        'label': ['204', 'A', '4', 'b', 'ii']})

        self.assertEqual([(1,), (2,), (5,)],
                         [c['sortable'] for c in parent['children']])

        parent = {'children': [
            {'node_type': 'appendix', 'label': ['204', 'A', '4', 'b']},
            {'node_type': 'appendix', 'label': ['204', 'A', '4', 'i']}
            ], 'label': ['204', 'appendix']}

        tree_builder.add_child(parent, {'node_type': 'appendix',
                                        'label': ['204', 'A', '4', 'g']})

        self.assertEqual([('b',), ('g',), ('i',)],
                         [c['sortable'] for c in parent['children']])



from regulations.generator.layers.base import LayerBase
from regulations.generator.layers.utils import is_contained_in


class FootnotesLayer(LayerBase):
    """Assembles the footnotes for this node, if available"""
    shorthand = 'footnotes'
    data_source = 'formatting'
    layer_type = LayerBase.PARAGRAPH

    def __init__(self, layer, version=None):
        self.layer = layer
        self.version = version

    def apply_layer(self, text_index):
        """
        Return a tuple of 'footnotes' and collection of footnotes.
        Footnotes are "collected" from the node and its children.
        .. note::
           This does not handle the case where the same note reference
           is used in multiple children.
        """
        footnotes = []
        for label in self.layer.keys():
            if is_contained_in(label, text_index):
                footnotes += [x['footnote_data']
                              for x in self.layer[label]
                              if 'footnote_data' in x]
        return 'footnotes', sorted(footnotes, key=lambda x: x['ref'])



from unittest import TestCase

from regulations.generator.label import Label


class LabelTests(TestCase):
    def test_sort_roman(self):
        """ Here we ensure that roman numerals are sorted correctly. """
        labels = [Label('200-20-d-2-viii'),
                  Label('200-20-d-2-ix'),
                  Label('200-20-d-2-iv'),
                  Label('200-20-d-2-v'),
                  Label('200-20-d-2-vi'),
                  Label('200-20-d-2-x'),
                  Label('200-20-d-2-xi')]
        labels = list(sorted(labels))
        self.assertEqual(labels, [Label('200-20-d-2-iv'),
                                  Label('200-20-d-2-v'),
                                  Label('200-20-d-2-vi'),
                                  Label('200-20-d-2-viii'),
                                  Label('200-20-d-2-ix'),
                                  Label('200-20-d-2-x'),
                                  Label('200-20-d-2-xi')])



from collections import defaultdict
from unittest import TestCase
from xml.etree import ElementTree as etree  # nosec - see usage below

from django.template import Context
from django.template.loader import get_template
from mock import Mock, patch

from regulations.generator.layers.formatting import FormattingLayer

template_loc = 'regulations/layers/{}.html'


class FormattingLayerTest(TestCase):
    def test_empty(self):
        """FormattingLayer ignores empty of missing node labels"""
        data = {'111-1': [], '111-2': [{'text': ''}]}
        with patch('regulations.generator.layers.formatting.loader') as ldr:
            render = ldr.get_template.return_value.render
            fl = FormattingLayer(data)
        self.assertEqual([], list(fl.apply_layer('111-0')))
        self.assertEqual([], list(fl.apply_layer('111-1')))
        self.assertEqual([], list(fl.apply_layer('111-2')))
        self.assertFalse(render.called)

    def assert_context_contains(self, template_name, data_key, data_value,
                                expected_context=None):
        """Verify that FormattingLayer converts the `data` elements into a
        context variable similar to `expect_context`. If `expected_context` is
        not provided, assume that it should match `data_value`"""
        if expected_context is None:
            expected_context = dict(data_value)
        data = {'111-3': [{'text': 'original', 'locations': [0, 2],
                           data_key: data_value}]}
        template_file = template_loc.format(template_name)
        with patch('regulations.generator.layers.formatting.loader') as ldr:
            # we will want to reference these templates later
            templates = defaultdict(Mock)
            ldr.get_template.side_effect = templates.__getitem__
            fl = FormattingLayer(data)
            result = list(fl.apply_layer('111-3'))
            render = templates[template_file].render

        self.assertEqual(len(result), 1)
        self.assertEqual('original', result[0][0])
        self.assertEqual([0, 2], result[0][2])

        self.assertTrue(render.called)
        context = render.call_args[0][0]
        for key, value in expected_context.items():
            self.assertTrue(key in context)
            self.assertEqual(context[key], value)

    def render_html(self, template_name, data):
        template_file = template_loc.format(template_name)
        template = get_template(template_file)
        return template.render(Context(data))

    def test_apply_layer_table(self):
        data = {'header': [[{'colspan': 2, 'rowspan': 1, 'text': 'Title'}]],
                'rows': [['cell 11', 'cell 12'], ['cell 21', 'cell 22']]}
        self.assert_context_contains('table', 'table_data', data)
        output = self.render_html('table', data)
        # Safe because: output comes from our own markup
        tree = etree.fromstring(output)     # nosec
        self.assertEqual(1, len(tree.findall(".//table/thead")))
        self.assertEqual(0, len(tree.findall(".//table/caption")))
        self.assertEqual('Title', tree.findall(".//table/thead/tr/th")[0].text)

    def test_apply_layer_table_with_caption(self):
        data = {'header': [[{'colspan': 2, 'rowspan': 1, 'text': 'Title'}]],
                'rows': [['cell 11', 'cell 12'], ['cell 21', 'cell 22']],
                'caption': 'Caption'}
        self.assert_context_contains('table', 'table_data', data)
        output = self.render_html('table', data)
        # Safe because: output comes from our own markup
        tree = etree.fromstring(output)     # nosec
        self.assertEqual(1, len(tree.findall(".//table/caption")))
        self.assertEqual('Caption', tree.findall(".//table/caption")[0].text)

    def test_apply_layer_note(self):
        data = {'type': 'note',
                'lines': ['Note:', '1. Content1', '2. Content2']}
        expected = {'lines': ['1. Content1', '2. Content2'], 'type': 'note'}
        self.assert_context_contains('note', 'fence_data', data, expected)

    def test_apply_layer_code(self):
        data = {'type': 'python',
                'lines': ['def double(x):', '    return x + x']}
        self.assert_context_contains('code', 'fence_data', data)

    def test_apply_layer_subscript(self):
        data = {'subscript': '123'}
        self.assert_context_contains('subscript', 'subscript_data', data)

    def test_apply_layer_superscript(self):
        data = {'superscript': '123'}
        self.assert_context_contains('superscript', 'superscript_data', data)

    def test_apply_layer_dash(self):
        data = {'text': 'This is an fp-dash'}
        self.assert_context_contains('dash', 'dash_data', data)

    def test_apply_layer_footnote(self):
        data = {'ref': '123', 'note': "Here's the note"}
        self.assert_context_contains('footnote', 'footnote_data', data)



import unittest

from regulations.uitests.base_test import BaseTest


class SubheaderTests(BaseTest, unittest.TestCase):
    job_name = 'Subheader tests'

    def test_effective_date(self):
        self.driver.get(self.test_url + '/1005-1/2012-12121')
        effective = self.driver.find_element_by_class_name('effective-date')
        self.assertIn('10/28/2012', effective.text)



import itertools
import logging


class AddQueue(object):
    """ Maintain a sorted list of nodes to add. This maintains a sorted queue
    of (label, node) tuples. """

    def __init__(self):
        self.queue = []

    def sort(self):
        self.queue = sorted(self.queue, key=lambda x: len(x[0]), reverse=True)

    def insert(self, item):
        self.queue.append(item)
        self.sort()

    def insert_all(self, items):
        self.queue += items
        self.sort()

    def find(self, label):
        found_nodes = [n for n in self.queue if n[0] == label]
        if found_nodes:
            return found_nodes[0]

    def delete(self, label):
        node_removed = [n for n in self.queue if n[0] != label]
        self.queue = node_removed


def build_label(node):
    return '-'.join(node['label'])


def build_tree_hash(tree):
    """ Build a hash map of a tree's nodes, so that we don't
    have to keep walking the tree. """
    tree_hash = {}

    def per_node(node):
        label_id = build_label(node)
        tree_hash[label_id] = node

        for c in node['children']:
            per_node(c)
    if tree:
        per_node(tree)
    return tree_hash


def parent_label(node):
    """This is not perfect. It can not handle children of subparts, for
    example"""
    if node['node_type'].upper() == 'INTERP':
        interpreting = list(itertools.takewhile(
            lambda l: l != 'Interp', node['label']))
        paragraph = node['label'][len(interpreting)+1:]
        if paragraph:
            return interpreting + ['Interp'] + paragraph[:-1]
        elif len(interpreting) == 1:    # Root of interpretations
            return interpreting
        else:
            return interpreting[:-1] + ['Interp']
    else:
        return node['label'][:-1]


def parent_in_tree(parent_label, tree_hash):
    """ Return True if the parent of node_label is in the tree """
    return parent_label in tree_hash


def add_node_to_tree(node, parent_label, tree_hash):
    """ Add the node to the tree by adding it to it's parent in order. """
    parent_node = tree_hash[parent_label]
    add_child(parent_node, node)
    return tree_hash


def roman_nums():
    """Generator for roman numerals."""
    mapping = [
        (1, 'i'), (4, 'iv'), (5, 'v'), (9, 'ix'),
        (10, 'x'), (40, 'xl'), (50, 'l'), (90, 'xc'),
        (100, 'c'), (400, 'cd'), (500, 'd'), (900, 'cm'),
        (1000, 'm')
        ]
    i = 1
    while True:
        next_str = ''
        remaining_int = i
        remaining_mapping = list(mapping)
        while remaining_mapping:
            (amount, chars) = remaining_mapping.pop()
            while remaining_int >= amount:
                next_str += chars
                remaining_int -= amount
        yield next_str
        i += 1


def make_label_sortable(label, roman=False):
    """ Make labels sortable, but converting them as appropriate.
    Also, appendices have labels that look like 30(a), we make those
    appropriately sortable. """

    if label.isdigit():
        return (int(label),)
    if roman:
        romans = list(itertools.islice(roman_nums(), 0, 50))
        return (1 + romans.index(label),)

    # segment the label piece into component parts
    # e.g. 45Ai33b becomes (45, 'A', 'i', 33, 'b')
    INT, UPPER, LOWER = 1, 2, 3
    segments, segment, seg_type = [], "", None
    for ch in label:
        if ch.isdigit():
            ch_type = INT
        elif ch.isalpha() and ch == ch.upper():
            ch_type = UPPER
        elif ch.isalpha() and ch == ch.lower():
            ch_type = LOWER
        else:
            # other character, e.g. parens, guarantee segmentation
            ch_type = None

        if ch_type != seg_type and segment:     # new type of character
            segments.append(segment)
            segment = ""

        seg_type = ch_type
        if ch_type:
            segment += ch

    if segment:    # ended with something other than a paren
        segments.append(segment)

    segments = [int(seg) if seg.isdigit() else seg for seg in segments]
    return tuple(segments)


def all_children_are_roman(parent_node):
    """
    Return true if all the children of the parent node have roman labels
    """
    romans = list(itertools.islice(roman_nums(), 0, 50))
    roman_children = [c['label'][-1] in romans
                      for c in parent_node['children']]
    return len(roman_children) > 0 and all(roman_children)


# @todo - this function is _very_ similar to one in regparser.notice.compiler.
# Can it be removed or pulled into a shared library?
def add_child(parent_node, node):
    "Add a child node to a parent, maintaining the order of the children."

    children = parent_node['children']
    children.append(node)
    child_labels = set('-'.join(c['label']) for c in children)
    order = parent_node.get('child_labels', [])

    if child_labels.issubset(set(order)):
        lookup = {'-'.join(c['label']): c for c in children}
        parent_node['children'] = [lookup[label_id] for label_id in order
                                   if label_id in child_labels]
    else:   # Explicit sort order not present/doesn't match nodes
        logging.warning(
            "No child_labels field. Guessing at child order (probably wrong)")
        for c in parent_node['children']:
            if c['node_type'].upper() == 'INTERP':
                if c['label'][-1] == 'Interp':
                    sortable = make_label_sortable(
                        c['label'][-2], roman=(len(c['label']) == 6))
                else:
                    paragraph = list(itertools.dropwhile(
                        lambda l: l != 'Interp', c['label']))[1:]
                    sortable = make_label_sortable(
                        paragraph[-1], roman=(len(paragraph) == 2))

                if len(parent_node['label']) == 2:
                    # Highest interpretation node in the land
                    p = len(list(itertools.takewhile(lambda l: l != 'Interp',
                                                     c['label'])))
                    prefix_length = (p, )
                    sortable = prefix_length + sortable
                c['sortable'] = sortable
            elif c['node_type'].upper() == 'APPENDIX':
                roman_children = all_children_are_roman(parent_node)
                c['sortable'] = make_label_sortable(c['label'][-1],
                                                    roman=roman_children)

            else:
                c['sortable'] = make_label_sortable(
                    c['label'][-1], roman=(len(c['label']) == 5))

        parent_node['children'].sort(key=lambda x: x['sortable'])



from collections import namedtuple

import six

from regulations.generator.label import Label
from regulations.generator.node_types import label_to_text
from regulations.generator.sidebar.base import SidebarBase


class SxSEntry(namedtuple('SxSEntry', ['label', 'doc_number'])):
    def template_context(self):
        return {'doc_number': self.doc_number,
                'label_id': self.label.id,
                'text': label_to_text(self.label.parts, include_section=False,
                                      include_marker=True)}


class Analyses(SidebarBase):
    """Section-by-Section analyses information, which describes the reasons
    for _why_ a section/paragraph has been modified."""
    shorthand = 'analyses'

    def context(self, http_client, request):
        analyses = []
        data = self.fetch_data(http_client)
        if data:
            tree_labels = [Label(parts=t['label'])
                           for t in self.fetch_relevant_trees(http_client)]
            analyses.extend(
                sxs for sxs in data
                if any(sxs.label in label for label in tree_labels))
        analyses = [analysis.template_context()
                    for analysis in sorted(analyses)]
        return {
            'analyses': analyses,
            'human_label_id': label_to_text(
                self.label_parts, include_marker=True),
            'version': self.version
        }

    def fetch_data(self, http_client):
        """Retrieves SxSEntry data from the server"""
        if 'Interp' in self.label_id:
            reg_part = self.label_parts[0]
            doc_id = reg_part + '-Interp'
        else:
            doc_id = self.label_id
        data = http_client.layer('analyses', 'cfr', doc_id, self.version)

        if not data:
            return []

        return [
            SxSEntry(Label(label_id), doc_number=subdata[-1]['reference'][0])
            for label_id, subdata in six.iteritems(data)]



# vim: set encoding=utf-8
from unittest import TestCase

from regulations.generator import node_types


class NodeTypesTest(TestCase):
    def test_change_appendix(self):
        node_parts_before = ['243', 'A', '30(a)']
        node_parts_after = node_types.to_markup_id(node_parts_before)
        node_string = "-".join(node_parts_after)

        self.assertEqual('243-A-30a', node_string)

    def test_type_from_label(self):
        for label in [('200', '5', 'A'), ('250',), ('250', '5'),
                      ('200', '5', 'a', 'i', 'C')]:
            self.assertEqual(node_types.REGTEXT,
                             node_types.type_from_label(label))

        for label in [('250', 'A2'), ('250', 'A'), ('250', 'A', '3(b)')]:
            self.assertEqual(node_types.APPENDIX,
                             node_types.type_from_label(label))

        for label in [('250', 'Interp'), ('250', 'A', 'Interp'),
                      ('250', 'A', 'Interp'), ('250', '5', 'b', 'Interp'),
                      ('250', '5', 'b', 'Interp', '1'),
                      ('250', '5', 'Interp', '5', 'r')]:
            self.assertEqual(node_types.INTERP,
                             node_types.type_from_label(label))

        self.assertEqual(node_types.EMPTYPART,
                         node_types.type_from_label(['250', 'Subpart']))
        self.assertEqual(node_types.SUBPART,
                         node_types.type_from_label(['250', 'Subpart', 'C']))

    def test_label_to_text(self):
        expectations = [
            (['2323', '4'], '2323.4'),
            (['2323', '5', 'r', '3'], '2323.5(r)(3)'),
            (['23', '5', 'r', '3', 'i', 'p12', 'A'], '23.5(r)(3)(i)'),
            (['23', '5', 'p1', 'a'], '23.5'),
            (['2323', 'A'], 'Appendix A to Part 2323'),
            (['2323', 'A', '4'], 'Appendix A-4'),
            (['2323', 'A', '4', 'b', '2'], 'Appendix A-4(b)(2)'),
            (['2323', '5', 'Interp'], 'Supplement to 2323.5'),
            (['2323', '7', 'b', 'Interp', '1', 'v'],
                'Supplement to 2323.7(b)-1.v'),
            (['2323', 'Z', 'Interp'], 'Supplement to Appendix Z to Part 2323'),
            (['204'], 'Regulation 204'),
            (['204', 'Interp'], 'Supplement I to Part 204'),
            (['204', 'Subpart', 'Interp'],
                'Interpretations for Regulation Text of Part 204'),
            (['204', 'Subpart', 'C', 'Interp'],
                'Interpretations for Subpart C of Part 204'),
            (['204', 'Appendices', 'Interp'],
                'Interpretations for Appendices of Part 204'),
            (['204', 'Interp', 'h1'], 'This Section'),
            (['204', 'M2'], 'Appendix M2 to Part 204')]
        for label, expected_text in expectations:
            self.assertEqual(expected_text, node_types.label_to_text(label))

        self.assertEqual('4', node_types.label_to_text(['2323', '4'], False))
        self.assertEqual(
            '5(r)(3)',
            node_types.label_to_text(['2323', '5', 'r', '3'], False))
        self.assertEqual(u'§ 2323.1',
                         node_types.label_to_text(['2323', '1'], True, True))
        self.assertEqual(u'§ 1',
                         node_types.label_to_text(['2323', '1'], False, True))



from datetime import datetime

from regulations.generator import api_reader
from regulations.generator.layers.utils import convert_to_python


def fetch_regulations_and_future_versions():
    """ Returns a dict for all the regulations in the API. The dict includes
    lists of future versions for each regulation. """
    client = api_reader.ApiReader()
    all_versions = client.all_regulations_versions()
    all_versions = convert_to_python(all_versions)

    regulations_future = {}

    # We're only interested in future endpoint versions
    for v in all_versions['versions']:
        if v['regulation'] not in regulations_future:
            regulations_future[v['regulation']] = []
        if 'by_date' in v:
            regulations_future[v['regulation']].append(v)
    return regulations_future


def fetch_grouped_history(part):
    client = api_reader.ApiReader()
    versions = [
        version for version in client.regversions(part)['versions']
        if 'by_date' in version
    ]
    for version in versions:
        version['notices'] = []
    versions = sorted(convert_to_python(versions), reverse=True,
                      key=lambda v: v['by_date'])

    today = datetime.today()
    seen_current = False

    for version in versions:
        if version['by_date'] > today:
            version['timeline'] = 'future'
        elif not seen_current:
            seen_current = True
            version['timeline'] = 'current'
        else:
            version['timeline'] = 'past'

    for notice in client.notices(part)['results']:
        notice = convert_to_python(notice)
        for version in versions:
            if version['by_date'] == notice.get('effective_on'):
                version['notices'].append(notice)

    for version in versions:
        version['notices'] = sorted(version['notices'], reverse=True,
                                    key=lambda n: n['publication_date'])

    return versions



from unittest import TestCase
from xml.etree import ElementTree as etree  # nosec - see usage below

from django.template import Context, Template


class MacrosTests(TestCase):
    def _gen_link(self, content):
        """Shorthand for passing the content into a template and rendering"""
        text = "{% load macros %}" + content
        as_str = Template(text).render(Context({}))
        # Safe because: we've constructed the XML
        as_xml = etree.fromstring("<ROOT>{}</ROOT>".format(as_str))  # nosec
        anchors = as_xml.findall('.//a')
        self.assertTrue(len(anchors) > 0)
        return anchors[0]

    def test_external_link_no_optional(self):
        """The classes and title fields are optional. We should generate an
        appropriate link"""
        anchor = self._gen_link(
            '{% external_link url="http://example.com/path" text="Click" %}')
        self.assertEqual(anchor.get('target'), '_blank')
        self.assertEqual(anchor.get('href'), 'http://example.com/path')
        self.assertFalse('title' in anchor.attrib)
        self.assertTrue('aria-label' in anchor.attrib)
        self.assertTrue('Click' in anchor.text)

    def test_external_link_classes_title(self):
        """The classes and title fields _can_ be added"""
        anchor = self._gen_link(
            '{% external_link url="url" text="text" classes="some here" '
            'title="My Title" %}')
        self.assertEqual(anchor.get('title'), 'My Title')
        self.assertTrue('some here' in anchor.get('class'))

    def test_search_for(self):
        """Macro should url-encode search terms."""
        anchor = self._gen_link(
            '{% search_for terms="has spaces" reg="1234" version="vvv" %}')
        self.assertTrue('1234' in anchor.get('href'))
        self.assertTrue('vvv' in anchor.get('href'))
        self.assertTrue('has%20spaces' in anchor.get('href'))



from unittest import TestCase

from regulations.generator.layers.definitions import DefinitionsLayer


class DefinitionsLayerTest(TestCase):
    def test_replacement_for(self):
        layer = {
            '202-3': {'ref': 'account:202-2-a'},
            'referenced': {'account:202-2-a': {
                'reference': '202-2-a', 'term': 'account'}}
            }
        dl = DefinitionsLayer(layer)
        definition_link = dl.replacement_for('account', layer['202-3'])

        url = '<a href="#202-2-a" class="citation definition" '
        url += 'data-definition="202-2-a" data-defined-term="account">'
        url += 'account</a>'
        self.assertEquals(definition_link, url)



from unittest import TestCase

from mock import patch

from regulations.generator.layers.paragraph_markers import (
    MarkerHidingLayer, MarkerInfoLayer, ParagraphMarkersLayer)


class ParagraphMarkersLayerTest(TestCase):
    @patch('regulations.generator.layers.paragraph_markers.loader')
    def test_apply_layer(self, loader):
        pml = ParagraphMarkersLayer({
            '1001-12-a': [{'text': '(a)', 'locations': [0]}],
            '1001-12-q': [{'text': 'q.', 'locations': [1]}]
        })
        self.assertEqual([], list(pml.apply_layer('1002-01-01')))

        a = list(pml.apply_layer('1001-12-a'))
        self.assertEqual(1, len(a))
        self.assertEqual('(a)', a[0][0])
        self.assertEqual([0], a[0][2])
        call_args = loader.get_template.return_value.render.call_args[0][0]
        self.assertEqual('(a)', call_args['paragraph'])
        self.assertEqual('a', call_args['paragraph_stripped'])

        q = list(pml.apply_layer('1001-12-q'))
        self.assertEqual(1, len(q))
        self.assertEqual('q.', q[0][0])
        self.assertEqual([1], q[0][2])
        call_args = loader.get_template.return_value.render.call_args[0][0]
        self.assertEqual('q.', call_args['paragraph'])
        self.assertEqual('q', call_args['paragraph_stripped'])


class MarkerHidingLayerTest(TestCase):
    @patch('regulations.generator.layers.paragraph_markers.loader')
    def test_apply_layer(self, loader):
        pml = MarkerHidingLayer({
            '1001-12-a': [{'text': '(a)', 'locations': [0]}],
            '1001-12-q': [{'text': 'q.', 'locations': [1]}]
        })
        self.assertEqual([], list(pml.apply_layer('1002-01-01')))

        a = list(pml.apply_layer('1001-12-a'))
        self.assertEqual(1, len(a))
        self.assertEqual('(a)', a[0][0])
        self.assertEqual([0], a[0][2])
        call_args = loader.get_template.return_value.render.call_args[0][0]
        self.assertEqual('(a)', call_args['paragraph'])

        q = list(pml.apply_layer('1001-12-q'))
        self.assertEqual(1, len(q))
        self.assertEqual('q.', q[0][0])
        self.assertEqual([1], q[0][2])
        call_args = loader.get_template.return_value.render.call_args[0][0]
        self.assertEqual('q.', call_args['paragraph'])


class MarkerInfoLayerTest(TestCase):
    def test_apply_layer(self):
        pml = MarkerInfoLayer({
            '1001-12-a': [{'text': '(a)', 'locations': [0]}],
            '1001-12-q': [{'text': 'q.', 'locations': [1]}]
        })
        self.assertIsNone(pml.apply_layer('1002-01-01'))
        self.assertEqual(tuple(pml.apply_layer('1001-12-a')),
                         ('paragraph_marker', 'a'))
        self.assertEqual(tuple(pml.apply_layer('1001-12-q')),
                         ('paragraph_marker', 'q'))



from django.http import HttpResponse
from django.template import RequestContext
from django.template.loader import select_template


def about(request):
    context = {}
    c = RequestContext(request, context)
    t = select_template([
        'regulations/custom-about.html',
        'regulations/about.html'])
    return HttpResponse(t.render(c))



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from collections import namedtuple
import logging
import re

from django.core.urlresolvers import reverse

from regulations.generator.toc import fetch_toc
from regulations.views import utils


logger = logging.getLogger(__name__)


class Title(namedtuple('Title', ['full', 'short', 'subtitle'])):
    def __new__(cls, full, short=None, subtitle=None):
        """Adds defaults to constructor"""
        return super(Title, cls).__new__(cls, full, short, subtitle)


class NavItem(namedtuple(
        'NavItem',
        ['url', 'title', 'markup_id', 'children', 'category', 'section_id'])):
    """Shared data structure to represent entries in the table of contents and
    the navigation in the page footer. We may be able to expand this
    standardization more broadly than fr_notices, but let's move one step at a
    time.
    :type title: Title
    :type markup_id: str
    :type children: potentially empty list
    :type category: str or None
    :param str section_id: markup id associated with AJAX and other JS;
        temporary shim so that we can turn off AJAX per NavItem. Defaults to
        copy the markup_id
    """
    def __new__(cls, url, title, markup_id, children=None, category=None,
                section_id=None):
        """Adds defaults to constructor"""
        if children is None:
            children = []
        if section_id is None:
            section_id = markup_id

        return super(NavItem, cls).__new__(
            cls, url, title, markup_id, children, category, section_id)

    # Properties/fns for backwards compatibility

    @property
    def markup_prefix(self):
        return self.title.short

    @property
    def sub_label(self):
        return self.title.subtitle


def _preamble_titles(node):
    """Hack: Split out navigation title and subtitle from a preamble node.
       TODO: Emit these fields in a ToC layer in -parser instead
       :param node: a preamble Node (i.e. dict)
       :return: pair of (title, sub_title) strings"""
    marker = node['label'][-1]
    prefix = '{}. '.format(marker.lower())
    normalized_title = node['title'].lower()
    if normalized_title.startswith(prefix):
        title, subtitle = node['title'].split('. ', 1)
        return Title(node['title'], title, subtitle)
    else:
        return Title(node['title'], marker, node['title'])


def make_preamble_nav(nodes, depth=1, max_depth=3):
    """Generate NavItems specific to a notice's preamble.
    :type nodes: iterable of Node (a dict)
    :param int depth: Current nesting depth of navigation elements
    :param int max_depth: We'll stop processing once we reach a certain depth
    """
    toc = []
    have_titles = [n for n in nodes if n.get('title')]
    for node in have_titles:
        url = reverse('chrome_preamble',
                      kwargs={'paragraphs': '/'.join(node['label'][:2])})
        # Add a hash to a specific section if we're not linking to the
        # top-level entry
        if len(node['label']) > 2:
            url += '#' + '-'.join(node['label'])

        markup_id = '{}-preamble-{}'.format(node['label'][0],
                                            '-'.join(node['label']))

        if 'intro' in node['label'] or depth == max_depth:
            children = []
        else:
            children = make_preamble_nav(
                node.get('children', []),
                depth=depth + 1,
                max_depth=max_depth)

        toc.append(NavItem(
            url=url,
            title=_preamble_titles(node),
            markup_id=markup_id,
            children=children
        ))
    return toc


class CFRChangeBuilder(object):
    """Builds the ToC specific to CFR changes from amendment data. As there is
    some valuable state shared between amendment processing, we store it all
    in an object"""
    def __init__(self):
        """version_info structure: {cfr_part -> {"left": str, "right": str}}
        e.g.  {"111": {"left": "v1", "right": "v2"},
               "222": {"left": "vold", "right": "vnew"}}"""
        self.cfr_title = self.cfr_part = self.section = None
        self.section_titles = {}
        self.toc = []

    def add_cfr_part(self, doc_number, version_info, amendment):
        """While processing an amendment, if it refers to a CFR part which
        hasn't been seen before, we need to perform some accounting, fetching
        related meta data, etc."""
        part = amendment['cfr_part']
        if part not in version_info:
            logger.warning("No version info for %s", part)
        elif self.cfr_part is None or self.cfr_part != amendment['cfr_part']:
            meta = utils.regulation_meta(part, version_info[part]['right'])
            flat_toc = fetch_toc(part, version_info[part]['right'],
                                 flatten=True)
            self.section_titles = {
                elt['index'][1]: elt['title']
                for elt in flat_toc if len(elt['index']) == 2}
            self.cfr_part = part
            self.cfr_title = meta.get('cfr_title_number')
            self.section = None

            title = '{} CFR {}'.format(self.cfr_title, part)
            markup_id = '{}-cfr-{}'.format(doc_number, part)
            self.toc.append(NavItem(
                url=reverse('cfr_changes', kwargs={
                    'doc_number': doc_number, 'section': part}),
                title=Title('Authority', title, 'Authority'),
                markup_id=markup_id,
                category=title,
                section_id=''))     # disable AJAX

    _cfr_re = re.compile(r'(§ [\d.]+) (.*)')

    def _change_title(self, section):
        title_str = self.section_titles.get(section)
        # Hack: Reconstitute node prefix and title
        # TODO: Emit these fields in a ToC layer in -parser instead
        match = self._cfr_re.search(title_str)
        if match:
            return Title(title_str, *match.groups())
        else:
            return Title(title_str, title_str)

    def add_change(self, doc_number, label_parts):
        """While processing an amendment, we will encounter sections we
        haven't seen before -- these will ultimately be ToC entries"""
        change_section = label_parts[1]
        is_subpart = 'Subpart' in label_parts or 'Subjgrp' in label_parts
        if not is_subpart and (self.section is None or
                               self.section != change_section):
            self.section = change_section
            section = '-'.join(label_parts[:2])

            self.toc.append(NavItem(
                url=reverse('cfr_changes', kwargs={
                    'doc_number': doc_number,
                    'section': section}),
                title=self._change_title(change_section),
                markup_id='{}-cfr-{}'.format(doc_number, section),
                category='{} CFR {}'.format(self.cfr_title, self.cfr_part)
            ))


def make_cfr_change_nav(doc_number, version_info, amendments):
    """Soup to nuts conversion from a document number to a table of contents
    list"""
    builder = CFRChangeBuilder()
    for amendment in amendments:
        # Amendments are of the form
        # {'cfr_part': 111, 'instruction': 'text1', 'authority': 'text2'} or
        # {'cfr_part': 111, 'instruction': 'text3',
        #  'changes': [['111-22-c', [data1]], ['other', [data2]]}
        builder.add_cfr_part(doc_number, version_info, amendment)
        for change_label, _ in amendment.get('changes', []):
            builder.add_change(doc_number, change_label.split('-'))
    return builder.toc


def footer(preamble_toc, cfr_toc, full_id):
    """Generate "navigation" context which allows the user to move between
    sections in the footer"""
    items = preamble_toc + cfr_toc
    nav = {'previous': None, 'next': None, 'page_type': 'preamble-section'}
    for idx, item in enumerate(items):
        if item.markup_id == full_id:
            if idx > 0:
                nav['previous'] = items[idx - 1]
            if idx < len(items) - 1:
                nav['next'] = items[idx + 1]
    return nav



# vim: set fileencoding=utf-8
from regulations.generator import title_parsing
from regulations.generator.layers.base import LayerBase
from regulations.generator.section_url import SectionUrl
from regulations.generator.toc import (
    toc_interp, toc_sect_appendix, toc_subpart)


class TableOfContentsLayer(LayerBase):
    shorthand = 'toc'
    data_source = 'toc'
    layer_type = LayerBase.PARAGRAPH

    def __init__(self, layer):
        self.layer = layer
        self.sectional = False
        self.version = None
        self.section_url = SectionUrl()

    def apply_layer(self, text_index):
        if text_index in self.layer:
            layer_elements = self.layer[text_index]

            toc_list = []
            for data in layer_elements:
                if 'Subpart' in data['index']:
                    toc_list.append(toc_subpart(data, toc_list, self.layer))
                elif 'Interp' in data['index']:
                    toc_list.append(toc_interp(data, toc_list, self.layer))
                else:
                    toc_list.append(toc_sect_appendix(data, toc_list))

            for el in toc_list:
                el['url'] = self.section_url.fetch(
                    el['index'], self.version, self.sectional)
                for sub in el.get('sub_toc', []):
                    sub['url'] = self.section_url.fetch(
                        sub['index'], self.version, self.sectional)
            return ('TOC', toc_list)

    @staticmethod
    def section(element, data):
        title_data = title_parsing.section(data)
        if title_data:
            element.update(title_data)

    @staticmethod
    def appendix_supplement(element, data, seen_appendix=False):
        as_data = title_parsing.appendix_supplement(data)
        if as_data:
            element.update(as_data)
        if element.get('is_appendix'):
            element['is_first_appendix'] = not seen_appendix



import unittest

from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support.select import Select

from regulations.uitests.base_test import BaseTest


class DiffTest(BaseTest, unittest.TestCase):

    job_name = 'Diff test'

    def get_drawer_button(self):
        return self.driver.find_element_by_xpath('//*[@id="timeline-link"]')

    def test_diffs(self):
        self.driver.get(self.test_url + '/1005-2/2011-11111')
        html = self.driver.find_element_by_tag_name('html')
        WebDriverWait(self.driver, 60).until(
            lambda driver: 'selenium-start' in html.get_attribute('class'))

        WebDriverWait(self.driver, 60)
        drawer_button = self.get_drawer_button()
        drawer_button.click()

        WebDriverWait(self.driver, 60).until(
            lambda driver: driver.find_element_by_css_selector(
                '#timeline:not(.hidden)'))

        # drawer button should be active
        self.assertIn('current', drawer_button.get_attribute('class'))

        diff_field = self.driver.find_element_by_css_selector(
            '#timeline .status-list:nth-child(2) form select')
        # select version to compare to
        Select(diff_field).select_by_value('2012-12121')
        diff_field.submit()

        # wait until diff view has loaded
        html = self.driver.find_element_by_tag_name('html')
        WebDriverWait(self.driver, 90).until(
            lambda driver: 'selenium-start' in html.get_attribute('class'))

        WebDriverWait(self.driver, 60)
        # make sure the url is right
        self.assertTrue(self.driver.current_url.startswith(
            self.test_url
            + '/diff/1005-2/2012-12121/2011-11111?from_version=2011-11111'),
            self.driver.current_url)

        WebDriverWait(self.driver, 60)

        # open diff pane in drawer
        active_drawer_button = self.get_drawer_button()
        active_drawer_button.click()
        WebDriverWait(self.driver, 60).until(
            lambda driver: 'current' in active_drawer_button.get_attribute(
                'class'))

        # navigate to 1005.3
        menu_link = self.driver.find_element_by_id('menu-link')
        WebDriverWait(self.driver, 10).until(
            lambda driver: menu_link.is_enabled() and menu_link.is_displayed())
        menu_link.click()
        WebDriverWait(self.driver, 10).until(
            lambda driver: 'current' in driver.find_element_by_id(
                'table-of-contents').get_attribute('class'))
        toc_link_1005_3 = self.driver.find_element_by_xpath(
            '//*[@id="toc"]/ol/li[3]')
        # drawer should have green bar
        self.assertTrue('modified' in toc_link_1005_3.get_attribute('class'))
        toc_link_1005_3.click()

        # wait until 1005.3 diff loads
        WebDriverWait(self.driver, 30).until(
            lambda driver:
            driver.current_url == self.test_url
            + '/diff/1005-3/2012-12121/2011-11111?from_version=2011-11111')

        # make sure new section is greened
        new_section = self.driver.find_element_by_xpath(
            '//*[@id="1005-3-b-1-vi"]')
        self.assertTrue(new_section.find_element_by_tag_name('ins'))

        # make sure changed paragraph has insertions and deletions
        changed_section = self.driver.find_element_by_xpath(
            '//*[@id="1005-3-b-2-ii"]')
        self.assertEqual(len(changed_section.find_elements_by_tag_name('ins')),
                         2)
        self.assertTrue(len(changed_section.find_elements_by_tag_name('del')))

        # go back into diff pane in drawer, stop comparing
        self.get_drawer_button().click()
        stop_button = self.driver.find_element_by_xpath(
            '//*[@id="timeline"]/div[2]/ul/li[2]/div/a')
        stop_button.click()

        # make sure it goes back to the right place
        WebDriverWait(self.driver, 30).until(
            lambda driver: driver.current_url.startswith(
                self.test_url + '/1005-3/2011-11111'))



# vim: set encoding=utf-8
from unittest import TestCase
from mock import patch
from regulations.views import navigation


class NavigationTest(TestCase):
    def test_get_labels(self):
        label = '204-1-a'
        self.assertEquals(
            ['204', '1', 'a'], navigation.get_labels(label))

    @patch('regulations.views.navigation.fetch_toc')
    def test_nav_sections(self, fetch_toc):
        fetch_toc.return_value = [
            {'index': ['204', '1'], 'title': '§ 204.1 First'},
            {'index': ['204', '3'], 'title': '§ 204.3 Third'}]
        nav = navigation.nav_sections('204-1', 'ver')
        p, n = nav
        self.assertEquals(None, p)
        self.assertEquals(['204', '3'], n['index'])
        self.assertEquals('§ 204.3 Third', n['title'])

    @patch('regulations.views.navigation.fetch_toc')
    @patch('regulations.views.navigation.SectionUrl')
    def test_nav_sections_prefix(self, su, fetch_toc):
        fetch_toc.return_value = [
            {'index': ['204', '1'], 'title': '§ 204.1 First',
             'is_section': True},
            {'index': ['204', 'A'], 'title': 'Appendix A'},
            {'index': ['204', 'Subpart', 'Interp'],
             'title': 'Regulation Text', 'is_subterp': True}]
        nav = navigation.nav_sections('204-A', 'ver')
        p, n = nav
        self.assertEqual(['204', '1'], p['index'])
        self.assertEqual('&sect;&nbsp;', p['markup_prefix'])
        self.assertEqual(['204', 'Subpart', 'Interp'], n['index'])
        self.assertEqual('Interpretations For ', n['markup_prefix'])

    @patch('regulations.views.navigation.fetch_toc')
    def test_nav_sections_appendix(self, fetch_toc):
        fetch_toc.return_value = [
            {'index': ['204', '1'], 'title': u'§ 204.1 First'},
            {'index': ['204', 'A'],
             'title': 'Appendix A to Part 204 - Model Forms'}]
        nav = navigation.nav_sections('204-1', 'ver')
        p, n = nav
        self.assertEquals(None, p)
        self.assertEquals(['204', 'A'], n['index'])
        self.assertEquals('Appendix A to Part 204 - Model Forms',
                          n['title'])

    @patch('regulations.views.navigation.fetch_toc')
    def test_nav_sections_subparts(self, fetch_toc):
        fetch_toc.return_value = [
            {'index': ['204', '1'], 'title': u'§ 204.1 First'},
            {'index': ['204', '2'], 'title': u'§ 204.2 Second'},
            {'index': ['204', '3'], 'title': u'§ 204.3 Third'},
            {'index': ['204', 'A'], 'title': 'Appendix A'},
            {'index': ['204', 'Subpart', 'A', 'Interp'], 'title': 'Subpart A'},
            {'index': ['204', 'Appendices', 'Interp'],
             'title': 'Appendices'}]
        nav = navigation.nav_sections('204-1', 'ver')
        p, n = nav
        self.assertEquals(None, p)
        self.assertEquals(['204', '2'], n['index'])

        nav = navigation.nav_sections('204-3', 'ver')
        p, n = nav
        self.assertEquals(['204', '2'], p['index'])
        self.assertEquals(['204', 'A'], n['index'])

        nav = navigation.nav_sections('204-A', 'ver')
        p, n = nav
        self.assertEquals(['204', '3'], p['index'])
        self.assertEquals(['204', 'Subpart', 'A', 'Interp'], n['index'])

        nav = navigation.nav_sections('204-Subpart-A-Interp', 'ver')
        p, n = nav
        self.assertEquals(['204', 'A'], p['index'])
        self.assertEquals(['204', 'Appendices', 'Interp'], n['index'])

        nav = navigation.nav_sections('204-Appendices-Interp', 'ver')
        p, n = nav
        self.assertEquals(['204', 'Subpart', 'A', 'Interp'], p['index'])
        self.assertEquals(None, n)



# vim: set encoding=utf-8
from unittest import TestCase

from mock import Mock

from regulations.generator.sidebar import analyses


class AnalysesSidebarTests(TestCase):
    def setUp(self):
        self.client = Mock()
        self.client.layer.return_value = {
            "111-22": [{"reference": ["2007-22", "111-22"],
                        "text": "Older analysis"},
                       {"reference": ["2009-11", "111-22"],
                        "text": "Newer analysis"}],
            "111-22-a": [{"reference": ["2009-22", "111-22-a"],
                          "text": "Paragraph analysis"}],
            '111-22-Interp': [{'reference': ['2007-22', '111-22-Interp']}],
            '111-22-Interp-2': [{'reference': ['2007-22', '111-22-Interp-2']}]
        }

    def test_context_section(self):
        self.client.regulation.return_value = {'label': ['111', '22']}
        self.assertEqual(
            analyses.Analyses('111-22', 'vvv').context(self.client, None),
            {'version': 'vvv',
             'human_label_id': u'§ 111.22',
             'analyses': [
                 {'doc_number': '2009-11', 'label_id': '111-22',
                  'text': u'§ 22'},
                 {'doc_number': '2009-22', 'label_id': '111-22-a',
                  'text': u'§ 22(a)'}
             ]})

    def test_context_paragraph(self):
        self.client.regulation.return_value = {"label": ['111', '22', 'a']}
        self.assertEqual(
            analyses.Analyses('111-22-a', 'vvv').context(self.client, None),
            {'version': 'vvv',
             'human_label_id': u'§ 111.22(a)',
             'analyses': [
                 {'doc_number': '2009-22', 'label_id': '111-22-a',
                  'text': u'§ 22(a)'}
             ]})

    def test_context_other_section(self):
        self.client.regulation.return_value = {"label": ['222', '22']}
        self.assertEqual(
            analyses.Analyses('222-22', 'vvv').context(self.client, None),
            {'version': 'vvv',
             'human_label_id': u'§ 222.22',
             'analyses': []})

    def test_context_interps(self):
        self.client.regulation.return_value = {
            "label": ['111', '22', 'Interp']}
        self.assertEqual(
            analyses.Analyses('111-22-Interp', 'vvv').context(self.client,
                                                              None),
            {'version': 'vvv',
             'human_label_id': 'Supplement to 111.22',
             'analyses': [
                 {'doc_number': '2007-22', 'label_id': '111-22-Interp',
                  'text': 'Supplement to 111.22'},
                 {'doc_number': '2007-22', 'label_id': '111-22-Interp-2',
                  'text': 'Supplement to 111.22-2'}
             ]})

    def test_context_interp_child(self):
        self.client.regulation.return_value = {
            "label": ['111', '22', 'Interp', '2']}
        self.assertEqual(
            analyses.Analyses('111-22-Interp-2', 'vvv').context(self.client,
                                                                None),
            {'version': 'vvv',
             'human_label_id': 'Supplement to 111.22-2',
             'analyses': [
                 {'doc_number': '2007-22', 'label_id': '111-22-Interp-2',
                  'text': 'Supplement to 111.22-2'}
             ]})



from importlib import import_module

import six
from django.conf import settings
from django.views.generic.base import TemplateView

from regulations.generator import api_reader


class SideBarView(TemplateView):
    """ View for handling the right-side sidebar """
    template_name = 'regulations/sidebar.html'
    # contains either class paths or class objects (not instances)
    components = settings.SIDEBARS

    def get_context_data(self, **kwargs):
        context = super(SideBarView, self).get_context_data(**kwargs)

        client = api_reader.ApiReader()

        klasses = []
        for class_or_class_path in self.components:
            if isinstance(class_or_class_path, six.string_types):
                module, class_name = class_or_class_path.rsplit('.', 1)
                klasses.append(getattr(import_module(module), class_name))
            else:
                klasses.append(class_or_class_path)

        sidebars = [klass(context['label_id'], context['version'])
                    for klass in klasses]
        context['sidebars'] = [sidebar.full_context(client, self.request)
                               for sidebar in sidebars]

        return context



from regulations.generator.layers.internal_citation import (
    InternalCitationLayer)
from mock import patch
from unittest import TestCase


class InternalCitationLayerTest(TestCase):

    @patch('regulations.generator.layers.internal_citation.loader')
    def test_render_url(self, loader):
        icl = InternalCitationLayer(None)

        icl.render_url(['888', '123'], 'link')
        context = loader.get_template.return_value.render.call_args[0][0]
        self.assertEqual('#888-123', context['citation']['url'])
        self.assertEqual('link', context['citation']['label'])

        icl.sectional = True
        icl.version = 'vvvv'
        icl.render_url(['888', '123'], 'look')
        context = loader.get_template.return_value.render.call_args[0][0]

        self.assertTrue('888-123/vvvv#888-123' in context['citation']['url'])
        self.assertEqual('look', context['citation']['label'])



from unittest import TestCase

from regulations.generator.layers.location_replace import LocationReplace


class LayersLocationReplaceTest(TestCase):
    def test_update_offsets_html(self):
        lr = LocationReplace()
        lr.update_offsets("a", "This is a test. It is only a test")
        self.assertEqual(lr.offsets, {0: (8, 9), 1: (27, 28)})
        lr.update_offsets("a", "This is a test. <a href='something'>link</a>")
        self.assertEqual(lr.offsets, {0: (8, 9)})

    def test_location_replace_text(self):
        lr = LocationReplace()
        replaced = lr.location_replace_text('Bunch of as as as',
                                            'as', '<sub>as</sub>', [0, 2])
        self.assertEqual('Bunch of <sub>as</sub> as <sub>as</sub>', replaced)

        lr = LocationReplace()
        replaced = lr.location_replace_text('Bunch of as as as',
                                            'as', '<sub>b</sub>', [0, 2])
        self.assertEqual('Bunch of <sub>b</sub> as <sub>b</sub>', replaced)

        lr = LocationReplace()
        replaced = lr.location_replace_text('Bunch of a_{s} a_{s} a_{s}',
                                            'a_{s}', 'a<sub>s</sub>', [0, 2])
        self.assertEqual('Bunch of a<sub>s</sub> a_{s} a<sub>s</sub>',
                         replaced)



from django.template import loader

from regulations.generator.layers import utils
from regulations.generator.layers.base import SearchReplaceLayer


class ExternalCitationLayer(SearchReplaceLayer):
    shorthand = 'external'
    data_source = 'external-citations'

    def __init__(self, layer):
        self.layer = layer
        self.template = loader.get_template(
            'regulations/layers/external_citation.html')

    def replacements_for(self, text, data):
        yield utils.render_template(self.template, data)



from django.http import HttpRequest

#   Don't import PartialInterpView or utils directly; causes an import cycle
from regulations import generator, views
from regulations.generator.layers.base import LayerBase
from regulations.generator.node_types import label_to_text
from regulations.generator.section_url import SectionUrl


class InterpretationsLayer(LayerBase):
    """Fetches the (rendered) interpretation for this node, if available"""
    shorthand = 'interp'
    data_source = 'interpretations'
    layer_type = LayerBase.PARAGRAPH

    def __init__(self, layer, version=None):
        self.layer = layer
        self.version = version
        self.section_url = SectionUrl()
        self.root_interp_label = None
        self.partial_view = None

    def preprocess_root(self, root):
        """The root label will allow us to use a single set of layer
        appliers and grab all interp data at once."""
        self.root_interp_label = '-'.join(root['label'] + ['Interp'])
        view_class = views.partial_interp.PartialInterpView
        self.partial_view = view_class.as_view(
            inline=True, appliers=view_class.mk_appliers(
                self.root_interp_label, self.version))

    def apply_layer(self, text_index):
        """Return a pair of field-name + interpretation if one applies."""
        if text_index in self.layer and self.layer[text_index]:
            context = {'interps': [],
                       'for_markup_id': text_index,
                       'for_label': label_to_text(text_index.split('-'),
                                                  include_section=False)}
            #   Force caching of a few nodes up -- should prevent a request
            #   per interpretation if caching is on
            generator.generator.get_tree_paragraph(
                self.root_interp_label, self.version)
            for layer_element in self.layer[text_index]:
                reference = layer_element['reference']

                request = HttpRequest()
                request.method = 'GET'
                response = self.partial_view(request, label_id=reference,
                                             version=self.version)
                response.render()

                interp = {
                    'label_id': reference,
                    'markup': response.content,
                }

                ref_parts = reference.split('-')
                interp['section_id'] = self.section_url.interp(
                    ref_parts, self.version)

                context['interps'].append(interp)

            return 'interp', context



from django.http import HttpResponse
from django.template import RequestContext
from django.template.loader import select_template

from regulations.generator import api_reader
from regulations.generator.versions import fetch_grouped_history
from regulations.views import utils


def regulation_exists(label_id):
    client = api_reader.ApiReader()
    vr = client.regversions(label_id)
    return (vr and len(vr) > 0)


def get_versions(label_id):
    """ Get the current and next version of the regulation. """
    history = fetch_grouped_history(label_id)
    if history:
        future = [h for h in history if h['timeline'] == 'future']
        if len(future) > 0:
            next_version = future[-1]
        else:
            next_version = None

        current = [h for h in history if h['timeline'] == 'current']
        current_version = current[0]
        return (current_version, next_version)


def regulation(request, label_id):

    context = {}
    current_version, new_version = get_versions(label_id)
    if new_version:
        context['new_version'] = new_version
    context['current_version'] = current_version

    context['label_id'] = label_id
    context['reg_first_section'] = utils.first_section(
        label_id, current_version['version'])
    context['reg_part'] = label_id.split('-')[0]

    context['meta'] = utils.regulation_meta(label_id,
                                            current_version['version'])

    c = RequestContext(request, context)

    t = select_template([
        'regulations/landing_%s.html' % label_id,
        'regulations/landing_base.html',
        'regulations/generic_landing.html'])
    return HttpResponse(t.render(c))



from unittest import TestCase

from mock import patch

from regulations.generator import subterp


class SubterpTest(TestCase):
    @patch('regulations.generator.subterp.fetch_toc')
    def test_filter_by_subterp(self, fetch_toc):
        nodes = [{'label': ['1005', 'h1', 'Interp']},
                 {'label': ['1005', '2', 'Interp']},
                 {'label': ['1005', '3', 'Interp']},
                 {'label': ['1005', '4', 'Interp']},
                 {'label': ['1005', 'A', 'Interp']},
                 {'label': ['1005', 'A_B', 'Interp']},
                 {'label': ['1005', 'B', 'Interp']}]
        self.assertEqual(nodes[1:4], subterp.filter_by_subterp(
            nodes, ['1005', 'Subpart', 'Interp'], 'vvvv'))
        self.assertFalse(fetch_toc.called)

        self.assertEqual(nodes[4:], subterp.filter_by_subterp(
            nodes, ['1005', 'Appendices', 'Interp'], 'vvvv'))
        self.assertFalse(fetch_toc.called)

        fetch_toc.return_value = [
            {'index': ['1005', 'Subpart', 'A'],
             'sub_toc': [{'index': ['1005', '1']}, {'index': ['1005', '2']}]},
            {'index': ['1005', 'Subpart', 'B'],
             'sub_toc': [{'index': ['1005', '3']}, {'index': ['1005', '4']}]},
            {'index': ['1005', 'A']},
            {'index': ['1005', 'B']},
            {'index': ['1005', 'Interp'],
             'sub_toc': [{'index': ['1005', 'Subpart', 'A', 'Interp']},
                         {'index': ['1005', 'Subpart', 'B', 'Interp']},
                         {'index': ['1005', 'Appendices', 'Interp']}]}]
        self.assertEqual(nodes[1:2], subterp.filter_by_subterp(
            nodes, ['1005', 'Subpart', 'A', 'Interp'], 'vvvv'))
        self.assertEqual(nodes[2:4], subterp.filter_by_subterp(
            nodes, ['1005', 'Subpart', 'B', 'Interp'], 'vvvv'))



from copy import deepcopy

from django.core.urlresolvers import reverse
from django.template.defaultfilters import title

from regulations.generator import api_reader, node_types
from regulations.generator.html_builder import PreambleHTMLBuilder
from regulations.generator.section_url import SectionUrl
from regulations.generator.versions import fetch_grouped_history
from regulations.views.partial import PartialView
import math

PAGE_SIZE = 10


url_rules = {
    'cfr': 'chrome_search',
    'preamble': 'chrome_search_preamble',
}


class PartialSearch(PartialView):
    """Display search results without any chrome."""
    template_name = 'regulations/search-results.html'

    def add_prev_next(self, current_page, context):
        total = float(context['results']['total_hits']) / PAGE_SIZE
        total = int(math.ceil(total))
        context['current'] = {'page': current_page + 1, 'total': total}

        if current_page > 0:
            context['previous'] = {'length': PAGE_SIZE,
                                   'page': current_page - 1}
        max_this_page = (current_page + 1) * PAGE_SIZE
        remaining = context['results']['total_hits'] - max_this_page
        if remaining > 0:
            context['next'] = {'page': current_page + 1,
                               'length': min(remaining, PAGE_SIZE)}

    def get_context_data(self, doc_type, **kwargs):
        # We don't want to run the content data of PartialView -- it assumes
        # we will be applying layers
        context = super(PartialView, self).get_context_data(**kwargs)

        context['q'] = self.request.GET.get('q')
        context['version'] = self.request.GET.get('version')
        context['doc_type'] = doc_type

        context['regulation'] = context['label_id'].split('-')[0]
        context['url_rule'] = url_rules[doc_type]

        try:
            page = int(self.request.GET.get('page', '0'))
        except ValueError:
            page = 0

        context['warnings'] = []
        if not context['q']:
            context['warnings'].append('Please provide a query.')
        if doc_type == 'cfr' and not context['version']:
            context['warnings'].append('Please provide a version.')

        if context['warnings']:
            results = {'results': [], 'total_hits': 0}
        else:
            results = api_reader.ApiReader().search(
                context['q'], context['doc_type'], context['version'],
                context['regulation'], page=page, is_root='false',
                is_subpart='false')

        if doc_type == 'cfr':
            context['results'] = process_cfr_results(results,
                                                     context['version'])
            for version in fetch_grouped_history(context['regulation']):
                for notice in version['notices']:
                    if notice['document_number'] == context['version']:
                        context['version_by_date'] = notice['effective_on']
        else:
            context['results'] = process_preamble_results(results)

        self.add_prev_next(page, context)

        return context


def process_cfr_results(results, version):
    """Modify the results of a search over the CFR by adding a human-readable
    label, appropriate links, and version information"""
    section_url = SectionUrl()
    results = deepcopy(results)
    for result in results['results']:
        result['header'] = node_types.label_to_text(result['label'])
        if 'title' in result:
            result['header'] += ' ' + title(result['title'])
        result['section_id'] = section_url.view_label_id(
            result['label'], version)
        result['url'] = section_url.fetch(
            result['label'], version, sectional=True)
    return results


def process_preamble_results(results):
    """Modify the results of a search over a notice preamble by adding a
    human-readable label, appropriate links, etc."""
    results = deepcopy(results)
    for result in results['results']:
        result['header'] = PreambleHTMLBuilder.human_label(result)
        if 'title' in result:
            result['header'] += ' ' + title(result['title'])
        result['section_id'] = '-'.join(
            [result['label'][0], 'preamble'] + result['label'])
        result['url'] = '{}#{}'.format(
            reverse(
                'chrome_preamble',
                kwargs={'paragraphs': '/'.join(result['label'][:2])},
            ),
            '-'.join(result['label']),
        )
    return results



# vim: set encoding=utf-8

from collections import namedtuple

import six
from django.core.urlresolvers import reverse

from regulations.generator import generator
from regulations.generator.html_builder import CFRHTMLBuilder
from regulations.generator.layers.toc_applier import TableOfContentsLayer
from regulations.generator.node_types import EMPTYPART, REGTEXT
from regulations.generator.section_url import SectionUrl
from regulations.generator.sidebar.diff_help import DiffHelp
from regulations.generator.toc import fetch_toc
from regulations.views import error_handling, utils
from regulations.views.chrome import ChromeView
from regulations.views.partial import PartialView


class Versions(namedtuple('Versions', ['older', 'newer', 'return_to'])):
    def __new__(cls, older, newer, return_to=None):
        if return_to is None:
            return_to = older
        return super(Versions, cls).__new__(cls, older, newer, return_to)


def get_appliers(label_id, versions):
    diff = generator.get_diff_applier(label_id, versions.older, versions.newer)

    if diff is None:
        raise error_handling.MissingContentException()

    layer_creator = generator.DiffLayerCreator(versions.newer)
    layer_creator.add_layers(
        ['graphics', 'paragraph', 'keyterms', 'defined', 'formatting',
         'marker-hiding', 'marker-info'],
        'cfr', label_id, version=versions.older)
    return layer_creator.get_appliers() + (diff, )


class PartialSectionDiffView(PartialView):
    """ A diff view of a partial section. """
    template_name = 'regulations/regulation-content.html'

    def get(self, request, *args, **kwargs):
        """ Override GET so that we can catch and propagate any errors. """

        try:
            return super(PartialSectionDiffView, self).get(request, *args,
                                                           **kwargs)
        except error_handling.MissingContentException:
            return error_handling.handle_generic_404(request)

    def footer_nav(self, label, toc, versions):
        nav = {}
        for idx, toc_entry in enumerate(toc):
            if toc_entry['section_id'] != label:
                continue

            if idx > 0:
                nav['previous'] = toc[idx - 1]

            if idx < len(toc) - 1:
                nav['next'] = toc[idx + 1]

        # Add the url
        for entry in nav.values():
            entry['url'] = reverse_chrome_diff_view(
                entry['section_id'], *versions)

        nav['page_type'] = 'diff'

        return nav

    def get_context_data(self, **kwargs):
        # We don't want to run the content data of PartialView -- it assumes
        # we will be applying layers
        context = super(PartialView, self).get_context_data(**kwargs)

        label_id = context['label_id']
        versions = Versions(context['version'], context['newer_version'],
                            self.request.GET.get('from_version'))

        tree = generator.get_tree_paragraph(label_id, versions.older)

        if tree is None:
            # TODO We need a more complicated check here to see if the diffs
            # add the requested section. If not -> 404
            tree = {}

        appliers = get_appliers(label_id, versions)

        builder = CFRHTMLBuilder(*appliers)
        builder.tree = tree
        builder.generate_html()

        child_of_root = builder.tree
        if builder.tree['node_type'] == REGTEXT:
            child_of_root = {
                'node_type': EMPTYPART,
                'children': [builder.tree]}
        context['tree'] = {'children': [child_of_root]}
        context['markup_page_type'] = 'diff'

        regpart = label_id.split('-')[0]
        old_toc = fetch_toc(regpart, versions.older)
        diff = generator.get_diff_json(regpart, versions.older, versions.newer)
        context['TOC'] = diff_toc(versions, old_toc, diff)
        context['navigation'] = self.footer_nav(
            label_id, context['TOC'], versions)
        return context


class ChromeSectionDiffView(ChromeView):
    """Search results with chrome"""
    template_name = 'regulations/diff-chrome.html'
    partial_class = PartialSectionDiffView
    sidebar_components = [DiffHelp]

    def check_tree(self, context):
        pass    # The tree may or may not exist in the particular version

    def add_diff_content(self, context):
        context['from_version'] = self.request.GET.get(
            'from_version', context['version'])
        context['left_version'] = context['version']
        context['right_version'] = \
            context['main_content_context']['newer_version']
        from_version = self.request.GET.get('from_version', context['version'])

        context['TOC'] = context['main_content_context']['TOC']

        #   Add reference to the first subterp, so we know how to redirect
        toc = fetch_toc(context['label_id'].split('-')[0], from_version)
        for entry in toc:
            if entry.get('is_supplement') and entry.get('sub_toc'):
                el = entry['sub_toc'][0]
                el['url'] = SectionUrl().of(
                    el['index'], from_version,
                    self.partial_class.sectional_links)
                context['first_subterp'] = el
        return context

    def add_main_content(self, context):
        super(ChromeSectionDiffView, self).add_main_content(context)
        return self.add_diff_content(context)


def reverse_chrome_diff_view(sect_id, left_ver, right_ver, from_version):
    """ Reverse the URL for a chromed diff view. """

    diff_url = reverse(
        'chrome_section_diff_view',
        args=(sect_id, left_ver, right_ver))
    diff_url += '?from_version=%s' % from_version
    return diff_url


def extract_sections(toc):
    compiled_toc = []
    for i in toc:
        if 'Subpart' in i['index'] or 'Subjgrp' in i['index']:
            compiled_toc.extend(i['sub_toc'])
        else:
            compiled_toc.append(i)
    return compiled_toc


def diff_toc(versions, old_toc, diff):
    # We work around Subparts in the TOC for now.
    compiled_toc = extract_sections(old_toc)

    for node in (v['node'] for v in diff.values() if v['op'] == 'added'):
        if len(node['label']) == 2 and node['title']:
            element = {
                'label': node['title'],
                'index': node['label'],
                'section_id': '-'.join(node['label']),
                'op': 'added'
            }
            data = {'index': node['label'], 'title': node['title']}
            TableOfContentsLayer.section(element, data)
            TableOfContentsLayer.appendix_supplement(element, data)
            compiled_toc.append(element)

    modified, deleted = modified_deleted_sections(diff)
    for el in compiled_toc:
        if 'Subpart' not in el['index'] and 'Subjgrp' not in el['index']:
            el['url'] = reverse_chrome_diff_view(el['section_id'], *versions)
        # Deleted first, lest deletions in paragraphs affect the section
        if tuple(el['index']) in deleted and 'op' not in el:
            el['op'] = 'deleted'
        if tuple(el['index']) in modified and 'op' not in el:
            el['op'] = 'modified'

    return sorted(compiled_toc, key=normalize_toc)


def normalize_toc(toc_element):
    """Return a sorting order for a TOC element, primarily based on the
    index, and the type of content. General order is regulation text,
    appendices, then interpretations."""

    sortable_index = tuple(utils.make_sortable(l)
                           for l in toc_element['index'])
    if toc_element.get('is_section'):
        return (0,) + sortable_index
    elif toc_element.get('is_appendix'):
        return (1,) + sortable_index
    elif toc_element.get('is_supplement'):
        return (2,) + sortable_index
    else:
        return (3,) + sortable_index


def modified_deleted_sections(diff):
    modified, deleted = set(), set()
    for label, diff_value in six.iteritems(diff):
        label = tuple(label.split('-'))
        if 'Interp' in label:
            section_label = (label[0], 'Interp')
        else:
            section_label = tuple(label[:2])

        # Whole section was deleted
        if diff_value['op'] == 'deleted' and label == section_label:
            deleted.add(section_label)
        # Whole section added/modified or paragraph added/deleted/modified
        else:
            modified.add(section_label)
    return modified, deleted



from setuptools import find_packages, setup


setup(
    name="regulations",
    version="6.0.2",
    packages=find_packages(),
    install_requires=[
        'boto3',
        'cached-property',
        'celery',
        'django>=1.8,<1.10',
        'enum34',
        'requests',
        'six',
        'requests-toolbelt',
    ],
    classifiers=[
        'License :: Public Domain',
        'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication'
    ]
)



import os
import shutil
import tempfile
from unittest import TestCase

from mock import patch

from regulations.apps import RegulationsConfig


class RegulationsConfigTests(TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    @patch('regulations.apps.get_app_template_dirs')
    def test_precompute_custom_templates(self, get_app_template_dirs):
        """Verify that custom templates are found"""
        get_app_template_dirs.return_value = [self.tmpdir]
        open(os.path.join(self.tmpdir, '123-45-a.html'), 'w').close()
        open(os.path.join(self.tmpdir, 'other.html'), 'w').close()

        RegulationsConfig.precompute_custom_templates()
        self.assertEqual(RegulationsConfig.custom_tpls['123-45-a'],
                         'regulations/custom_nodes/123-45-a.html')
        self.assertEqual(RegulationsConfig.custom_tpls['other'],
                         'regulations/custom_nodes/other.html')
        self.assertFalse('another' in RegulationsConfig.custom_tpls)

    @patch('regulations.apps.get_app_template_dirs')
    def test_precompute_node_type_templates(self, get_app_template_dirs):
        """Verify that node-type templates are found and that a default dict is
        used"""
        get_app_template_dirs.return_value = [self.tmpdir]
        open(os.path.join(self.tmpdir, 'a_new_type.html'), 'w').close()

        RegulationsConfig.precompute_node_type_templates()
        self.assertEqual(RegulationsConfig.node_type_tpls['extract'],
                         'regulations/tree/extract.html')
        self.assertEqual(RegulationsConfig.node_type_tpls['a_new_type'],
                         'regulations/tree/a_new_type.html')
        self.assertEqual(RegulationsConfig.node_type_tpls['something_else'],
                         'regulations/tree/default.html')



import re

from django.template import loader

from regulations.generator.layers import utils
from regulations.generator.layers.base import LayerBase, SearchReplaceLayer


class ParagraphMarkersLayer(SearchReplaceLayer):
    """This layer creates a new paragraph marker (with periods) and adds a
    class to the old (to make it easier to hide). It exists for backwards
    compatibility; new clients should be using MarkerHidingLayer and
    MarkerInfoLayer instead"""
    shorthand = 'paragraph'
    data_source = 'paragraph-markers'
    TO_STRIP_RE = re.compile(r'[().]')

    def __init__(self, layer):
        self.layer = layer
        self.template = loader.get_template(
            'regulations/layers/paragraph_markers.html')

    def replacements_for(self, original, data):
        stripped = self.TO_STRIP_RE.sub('', original)

        context = {'paragraph': original, 'paragraph_stripped': stripped}
        yield utils.render_template(self.template, context)


class MarkerHidingLayer(SearchReplaceLayer):
    """This layer wraps the original marker in a class, making it easier to
    hide."""
    shorthand = 'marker-hiding'
    data_source = 'paragraph-markers'

    def __init__(self, layer):
        self.layer = layer
        self.template = loader.get_template(
            'regulations/layers/marker_hiding.html')

    def replacements_for(self, original, data):
        context = {'paragraph': original}
        yield utils.render_template(self.template, context)


class MarkerInfoLayer(LayerBase):
    """This layer adds the paragraph marker as an attribute of the node. This
    is then used to display the marker outside of the normal position"""
    shorthand = 'marker-info'
    data_source = 'paragraph-markers'
    layer_type = LayerBase.PARAGRAPH

    def __init__(self, layer):
        self.layer_data = layer

    def apply_layer(self, text_index):
        if self.layer_data.get(text_index):
            original = self.layer_data[text_index][0]["text"]
            stripped = original.replace('(', '').replace(')', '')
            stripped = stripped.replace('.', '')
            return 'paragraph_marker', stripped



from unittest import TestCase
from django.template import loader, Context


class PartialDefinitionTests(TestCase):
    urls = 'regulation.urls'

    def test_partial_definition_with_children(self):
        t = loader.get_template('regulations/partial-definition.html')

        node = {
            'section_id': '102-4',
            'label_id': '202-2-a',
            'children': [{'label_id': '202-2-a-1'}]}

        context_dict = {'node': node, 'version': '2012-1223'}
        response = t.render(Context(context_dict))
        best_viewed = 'This definition is best viewed in its original location'
        self.assertTrue(best_viewed in response)

    def test_partial_definition_no_children(self):
        t = loader.get_template('regulations/partial-definition.html')

        node = {
            'section_id': '102-4',
            'label_id': '202-2-a',
            'marked_up': 'This term is defined carefully'}

        context_dict = {'node': node, 'version': '2012-1223'}
        response = t.render(Context(context_dict))
        carefully = 'This term is defined carefully'
        self.assertTrue(carefully in response)



from django.template import Library, Context, loader

register = Library()


@register.simple_tag()
def render_nested(template, context=None):
    return loader.get_template(template).render(Context(context))



from itertools import dropwhile, takewhile

from regulations.generator.toc import fetch_toc


def filter_by_subterp(nodes, subterp_label, version):
    """Given an interp node (e.g. a node with label ['1005', 'Interp']),
    remove all irrelevant sub-nodes. Note that stripping out nodes is more
    efficient than compiling nodes when API calls are involved. We use
    takewhile and dropwhile in case extra, non-node interpretations are in
    the range"""

    def is_section(n):
        return n['label'][1].isdigit()

    def not_section(n):
        return not is_section(n)

    if subterp_label[1:] == ['Subpart', 'Interp']:      # Empty part
        skip_intros = dropwhile(not_section, nodes)
        sections = takewhile(is_section, skip_intros)
        return list(sections)
    elif subterp_label[1:] == ['Appendices', 'Interp']:  # Appendices
        skip_intros = dropwhile(not_section, nodes)
        skip_sections = dropwhile(is_section, skip_intros)
        return list(skip_sections)
    else:   # A Subpart. Most costly as we need to know the toc
        subpart_label = subterp_label[:-1]
        toc = fetch_toc(subterp_label[0], version)
        toc = list(dropwhile(
            lambda el: el['index'] != subpart_label,    # not subpart
            toc))
        if toc:
            sections = set(tuple(el['index'] + ['Interp'])
                           for el in toc[0]['sub_toc'])
            relevant_interps = dropwhile(
                lambda el: tuple(el['label']) not in sections,   # not relevant
                nodes)

            other_sections = set()
            for el in toc[1:]:
                other_sections.add(tuple(el['index'] + ['Interp']))
                for sub in el.get('sub_toc', []):
                    other_sections.add(tuple(sub['index'] + ['Interp']))
            relevant_interps = takewhile(
                lambda n: tuple(n['label']) not in other_sections,  # relevant
                relevant_interps)

            return list(relevant_interps)
    # Couldn't find the subpart - wrong label? - implicit return None



# vim: set encoding=utf-8
import unittest

from nose.tools import assert_equal
from selenium.webdriver.support.ui import WebDriverWait

from regulations.uitests.base_test import BaseTest
from regulations.uitests import utils


class ScrollTest(BaseTest, unittest.TestCase):

    job_name = 'Scroll test'

    def test_scroll(self):
        self.driver.get(self.test_url + '/1005-36/2011-11111')

        header = self.driver.find_element_by_css_selector('.header-label')
        assert_equal(header.text, u'\xa7 1005.36')

        utils.scroll_to(self.driver, '#1005-36-a')
        WebDriverWait(self.driver, 5).until(
            lambda driver: header.text == u'\xa7 1005.36(a)')

        utils.scroll_to(self.driver, '#1005-36-a-2-iii')
        WebDriverWait(self.driver, 5).until(
            lambda driver: header.text == u'\xa7 1005.36(a)(2)(iii)')



# vim: set encoding=utf-8
from itertools import takewhile
import re

# These node types represent categories of paragraphs/nodes within the
# regulation tree.
# APPENDIX - Auxiliary material at the end of the regulation
APPENDIX = u'appendix'
# INTERP - A special type of appendix dedicated to agency interpretations of
# the rest of the regulation
INTERP = u'interp'
# REGTEXT - The most common type of node. This represents paragraphs and
# sections
REGTEXT = u'regtext'
# SUBPART - Regulations are often split into lettered groupings of sections;
# the parent node of all of these sections is a SUBPART
SUBPART = u'subpart'
# SUBJGRP - Less common, but very similar to SUBPARTs, SUBJGRPs are groupings
# of sections without a specific letter designation
SUBJGRP = u'subjgrp'
# EMPTYPART - This is a "virtual" node type in the sense that it does not
# correspond to anything in the original regulation. Instead, it wraps
# sections which do not live inside a SUBPART/SUBJGRP to provide parallelism
EMPTYPART = u'emptypart'

PAREN_RE = re.compile(r'[()]')


def to_markup_id(id_parts):
    """Given the id parts from the JSON tree, convert to an id that can
    be used in the front end"""
    new_id = list(id_parts)
    if type_from_label(id_parts) in (APPENDIX, INTERP):
        return [PAREN_RE.sub('', part) for part in new_id]
    return new_id


def type_from_label(label):
    """Given a list of label parts, determine the associated node's type"""
    if 'Interp' in label:
        return INTERP
    if label[-1] == 'Subpart':
        return EMPTYPART
    if 'Subpart' in label:  # but not the final segment
        return SUBPART
    if 'Subjgrp' in label:
        return SUBJGRP
    if len(label) > 1 and label[1][:1].isalpha():
        return APPENDIX
    return REGTEXT


def label_to_text(label, include_section=True, include_marker=False):
    """Convert a label:list[string] into a human-readable string"""
    if len(label) == 1:
        return 'Regulation %s' % label[0]

    # Use short circuiting to grab the *first* type of label that matches
    return (_l2t_subterp(label) or _l2t_interp(label) or _l2t_appendix(label)
            or _l2t_section(label, include_section, include_marker))


MARKERLESS_REGEX = re.compile(r'^[hp]\d+')


def _not_markerless(l):
    return not MARKERLESS_REGEX.match(l)


def take_until_markerless(label_parts):
    return list(takewhile(_not_markerless, label_parts))


def _join_paragraph_tail(label_parts, join_with, prefix='', suffix=''):
    """Given the tail of paragraph markers in a label, convert them into a
    string, separated by the appropriate strings (join_with). Also remove any
    markers following a markerless paragraph"""
    label_parts = take_until_markerless(label_parts)
    if label_parts:
        return prefix + join_with.join(label_parts) + suffix
    else:
        return ""


def _l2t_subterp(label):
    """Helper function converting subterp labels to text. Assumes label has
    more then one segment"""
    if label[1:] == ['Subpart', 'Interp']:
        return 'Interpretations for Regulation Text of Part ' + label[0]
    elif label[1:] == ['Appendices', 'Interp']:
        return 'Interpretations for Appendices of Part ' + label[0]
    elif len(label) == 4 and label[1] == 'Subpart' and label[3] == 'Interp':
        interpretations_for = 'Interpretations for Subpart '
        return interpretations_for + label[2] + ' of Part ' + label[0]


def _l2t_interp(label):
    """Helper function converting interpretation labels to text. Assumes
    _l2t_subterp failed"""
    if 'Interp' in label:
        # Interpretation
        prefix = list(takewhile(lambda l: l != 'Interp', label))
        suffix = label[label.index('Interp') + 1:]
        if len(prefix) == 1 and suffix:
            # Interpretation introduction; for now we cop out
            return 'This Section'
        elif len(prefix) == 1:
            return 'Supplement I to Part %s' % prefix[0]
        elif suffix:
            suffix = _join_paragraph_tail(suffix, '.')
            return 'Supplement to %s-%s' % (label_to_text(prefix), suffix)
        else:
            return 'Supplement to %s' % label_to_text(prefix)


def _l2t_appendix(label):
    """Helper function converting appendix labels to text. Assumes
    _l2t_subterp and _l2t_interp failed"""
    if type_from_label(label) == APPENDIX:
        # Appendix
        label = take_until_markerless(label)
        if len(label) == 2:  # e.g. 225-B
            return 'Appendix ' + label[1] + ' to Part ' + label[0]
        elif len(label) == 3:  # e.g. 225-B-3
            return 'Appendix %s-%s' % tuple(label[1:])
        else:  # e.g. 225-B-3-a-4-i
            suffix = _join_paragraph_tail(label[3:], ')(', '(', ')')
            return 'Appendix %s-%s%s' % (label[1], label[2], suffix)


def _l2t_section(label, include_section, include_marker):
    """Helper function converting section labels to text. Assumes
    _l2t_subterp, _l2t_interp, and _l2t_appendix failed"""
    if include_marker:
        marker = u'§ '
    else:
        marker = ''

    if include_section:
        # Regulation Text with section number
        if len(label) == 2:  # e.g. 225-2
            return marker + '.'.join(label)
        else:  # e.g. 225-2-b-4-i-A
            suffix = _join_paragraph_tail(label[2:], ')(', '(', ')')
            return marker + '%s.%s%s' % (label[0], label[1], suffix)
    else:
        # Regulation Text without section number
        if len(label) == 2:  # e.g. 225-2
            return marker + label[1]
        else:  # e.g. 225-2-b-4-i-A
            suffix = _join_paragraph_tail(label[2:], ')(', '(', ')')
            return marker + label[1] + suffix



from unittest import TestCase
from mock import Mock, patch

from django.test import RequestFactory

from regulations.generator.layers.layers_applier import (
    InlineLayersApplier, ParagraphLayersApplier, SearchReplaceLayersApplier)
from regulations.generator.node_types import EMPTYPART, INTERP, REGTEXT
from regulations.views import partial


class PartialParagraphViewTests(TestCase):
    @patch.object(partial.PartialParagraphView, 'section_navigation')
    def test_transform_context(self, sn):
        ppv = partial.PartialParagraphView()
        builder = Mock()
        builder.tree = {'label': ['1111', '22', 'a', '5'],
                        'node_type': REGTEXT}
        ctx = {'label_id': '1111-22-a-5-i', 'version': 'vvvv'}
        ppv.transform_context(ctx, builder)
        self.assertEqual(ctx['tree'], {
            'node_type': REGTEXT, 'label': ['1111'],
            'children': [{
                'node_type': EMPTYPART, 'label': ['1111', 'Subpart'],
                'children': [{
                    'node_type': REGTEXT, 'label': ['1111', '22'],
                    'markup_id': '1111-22-a-5-i',
                    'children': [{
                        'node_type': REGTEXT, 'label': ['1111', '22', 'a'],
                        'children': [builder.tree]}]}]}]})

        builder.tree = {'label': ['1111', 'Interp', 'h1'],
                        'node_type': INTERP}
        ctx = {'label_id': '1111-Interp-h1', 'version': 'vvvv'}
        ppv.transform_context(ctx, builder)
        self.assertEqual(ctx['tree'], {
            'node_type': REGTEXT, 'label': ['1111'],
            'children': [{
                'node_type': INTERP, 'label': ['1111', 'Interp'],
                'markup_id': '1111-Interp-h1',
                'children': [builder.tree]}]})


class PartialSectionViewTests(TestCase):
    @patch('regulations.generator.generator.get_tree_paragraph')
    @patch('regulations.views.partial.navigation')
    @patch('regulations.generator.generator.LayerCreator.get_layer_json')
    def test_get_context_data(self, get_layer_json, navigation,
                              get_tree_paragraph):
        get_layer_json.return_value = {'layer': 'layer'}
        navigation.nav_sections.return_value = None, None
        get_tree_paragraph.return_value = {
            'text': 'Some Text',
            'children': [],
            'label': ['205'],
            'node_type': REGTEXT
        }

        reg_part_section = '205'
        reg_version = '2013-10608'

        request = RequestFactory().get('/fake-path/?layers=meta')
        view = partial.PartialSectionView.as_view(
            template_name='regulations/regulation-content.html')

        response = view(request, label_id=reg_part_section,
                        version=reg_version)
        root = response.context_data['tree']
        subpart = root['children'][0]
        self.assertEqual(subpart['children'][0],
                         get_tree_paragraph.return_value)


class PartialViewTest(TestCase):

    def test_generate_html(self):
        regulation_tree = {'text': '', 'children': [], 'label': ['8675'],
                           'title': 'Regulation R', 'node_type': REGTEXT}
        i_applier = InlineLayersApplier()
        p_applier = ParagraphLayersApplier()
        sr_applier = SearchReplaceLayersApplier()
        appliers = (i_applier, p_applier, sr_applier)
        builder = partial.generate_html(regulation_tree, appliers)
        self.assertEquals(builder.tree, regulation_tree)
        self.assertEquals(builder.inline_applier, i_applier)
        self.assertEquals(builder.p_applier, p_applier)
        self.assertEquals(builder.search_applier, sr_applier)



from collections import defaultdict
import os

from django.apps import AppConfig
from django.template.loaders.app_directories import get_app_template_dirs


class RegulationsConfig(AppConfig):
    name = 'regulations'
    # Maps node label_id's to the template which should be used to render
    # them. Warning: global state
    custom_tpls = {}
    node_type_tpls = defaultdict(lambda: "regulations/tree/default.html")

    def ready(self):
        """Called (almost) once per application startup. Should only contain
        idempotent operations"""
        self.precompute_custom_templates()
        self.precompute_node_type_templates()

    @staticmethod
    def _find_templates(location, add_to):
        """Helper method for shared search-templates-add-to-dict
        functionality"""
        search_path = os.path.join('templates', 'regulations', location)
        file_names = {file_name
                      for template_dir in get_app_template_dirs(search_path)
                      for file_name in os.listdir(template_dir)
                      if os.path.isfile(os.path.join(template_dir, file_name))}
        for file_name in file_names:
            ident, _ = os.path.splitext(file_name)
            add_to[ident] = 'regulations/{}/{}'.format(location, file_name)

    @classmethod
    def precompute_custom_templates(cls):
        """We allow agencies to provide templates for specific nodes in the
        regulation tree. Rather than have the rendering code inspect which
        templates are available at render time, we'll find the special cases
        during the application startup"""
        cls._find_templates('custom_nodes', cls.custom_tpls)

    @classmethod
    def precompute_node_type_templates(cls):
        """Different node types may display differently. Scan the provided
        templates to see if any node-type-specific templates are present."""
        cls._find_templates('tree', cls.node_type_tpls)



from unittest import TestCase

from django.conf import settings
from django.http import HttpResponseGone
from django.test import Client, RequestFactory
from mock import patch

from regulations.views import chrome


class ViewsChromeTest(TestCase):
    def setUp(self):
        self.original_debug = settings.DEBUG
        self.original_api_base = settings.API_BASE

    def tearDown(self):
        settings.DEBUG = self.original_debug
        settings.API_BASE = self.original_api_base

    @patch('regulations.views.error_handling.api_reader')
    @patch('regulations.views.chrome.ChromeView.set_chrome_context')
    @patch('regulations.views.chrome.generator')
    def test_404(self, generator, set_chrome_context, api_reader):
        """Test that the response of the outer view is that of the inner
        when there's an error"""
        api_reader.ApiReader.return_value.regversions.return_value = None
        generator.get_tree_paragraph.return_value = {}
        set_chrome_context.return_value = None

        class InnerView(chrome.TemplateView):
            def get(self, request, *args, **kwargs):
                return HttpResponseGone()

        class FakeView(chrome.ChromeView):
            partial_class = InnerView

        view = FakeView()
        view.request = RequestFactory().get('/')
        response = view.get(view.request, label_id='lab', version='ver')
        self.assertEqual(404, response.status_code)

    @patch('regulations.views.chrome.error_handling')
    @patch('regulations.views.chrome.generator')
    @patch('regulations.views.chrome.SideBarView')
    def test_error_propagation(self, sbv, generator, error_handling):
        """While we don't rely on this sort of propagation for the main
        content (much), test it in the sidebar"""
        sbv.as_view.return_value.return_value = HttpResponseGone()

        class FakeView(chrome.ChromeView):
            def add_main_content(self, context):
                pass

            def set_chrome_context(self, context, reg_part, version):
                pass

        view = FakeView()
        view.request = RequestFactory().get('/')
        response = view.get(view.request, label_id='lab', version='ver')
        self.assertEqual(410, response.status_code)

    @patch('regulations.views.chrome.generator')
    def test_get_404(self, generator):
        generator.get_regulation.return_value = None
        response = Client().get('/regulation/111/222')
        self.assertEqual(404, response.status_code)

    @patch('regulations.views.chrome.generator')
    def test_get_404_tree(self, generator):
        generator.get_regulation.return_value = {'regulation': 'tree'}
        generator.get_tree_paragraph.return_value = None
        response = Client().get('/regulation/111/222')
        self.assertEqual(404, response.status_code)

    def test_diff_redirect_label_regulation(self):
        """If viewing a full regulation, the redirect for diffs should point
        to the first section"""
        view = chrome.ChromeView()
        toc = [{'section_id': '199-Subpart-A',
                'sub_toc': [{'section_id': '199-4'}, {'section_id': '199-6'}]},
               {'section_id': '199-Subpart-B',
                'sub_toc': [{'section_id': '199-8'}, {'section_id': '199-9'}]}]
        self.assertEqual('199-4', view.diff_redirect_label('199', toc))

    def test_diff_redirect_label_paragraph(self):
        """If viewing a single paragraph, the redirect for diffs should point
        to that paragraph's section. Similarly, all diffs for interpretations
        should point to the root interpretation"""
        view = chrome.ChromeView()
        self.assertEqual('199-4', view.diff_redirect_label('199-4-b', []))
        self.assertEqual('199-4', view.diff_redirect_label('199-4-b-3', []))
        self.assertEqual('199-A', view.diff_redirect_label('199-A', []))
        self.assertEqual('199-A', view.diff_redirect_label('199-A-14B', []))
        self.assertEqual('199-Interp',
                         view.diff_redirect_label('199-Interp', []))
        self.assertEqual('199-Interp',
                         view.diff_redirect_label('199-Interp-h1', []))
        self.assertEqual('199-Interp',
                         view.diff_redirect_label('199-12-Interp-2', []))


class ViewsChromeSubterpTest(TestCase):
    def test_diff_redirect_label(self):
        view = chrome.ChromeSubterpView()
        for label in ('199-Subpart-Interp', '199-Subpart-A-Interp',
                      '199-Appendices-Interp'):
            self.assertEqual('199-Interp',
                             view.diff_redirect_label(label, None))

    @patch('regulations.views.chrome.generator')
    @patch('regulations.views.chrome.filter_by_subterp')
    def test_check_tree(self, filter_by_subterp, generator):
        view = chrome.ChromeSubterpView()

        generator.get_tree_paragraph.return_value = None
        try:
            view.check_tree({'version': 'vvvv', 'label_id': 'llll'})
            self.assertTrue(False)
        except chrome.error_handling.MissingSectionException:
            pass

        generator.get_tree_paragraph.return_value = {'children': []}
        filter_by_subterp.return_value = []
        try:
            view.check_tree({'version': 'vvvv', 'label_id': 'llll'})
            self.assertTrue(False)
        except chrome.error_handling.MissingSectionException:
            pass

        filter_by_subterp.return_value = ["something"]
        view.check_tree({'version': 'vvvv', 'label_id': 'llll'})
        # No exception



from .base import *

DEBUG = True

CACHES['default']['BACKEND'] = 'django.core.cache.backends.dummy.DummyCache'
CACHES['eregs_longterm_cache']['BACKEND'] = 'django.core.cache.backends.dummy.DummyCache'
CACHES['api_cache']['TIMEOUT'] = 5  # roughly per request

INSTALLED_APPS += (
    'django_nose',
)

NOSE_ARGS = [
    '--with-coverage',
    '--cover-package=fr_notices,regulations',
    '--exclude-dir=regulations/uitests'
]

try:
    from local_settings import *
except ImportError:
    pass



# -*- coding: utf-8 -*-
from django.test import TestCase
from mock import patch
import httpretty
import requests

from regulations.docket import sanitize_fields


class SanitizeFieldsTest(TestCase):
    def setUp(self):
        self.patch_document_fields = patch(
            'regulations.docket.get_document_fields')
        mock_object = self.patch_document_fields.start()
        mock_object.return_value = {
            'required_field': {'maxLength': 10, 'required': True},
            'optional_field': {'maxLength': 10, 'required': False},
        }

    def tearDown(self):
        self.patch_document_fields.stop()

    def test_valid_body(self):
        test_body = {
            "required_field": "Value 1",
            "optional_field": "Value 2"
        }
        valid, message = sanitize_fields(test_body)
        self.assertTrue(valid)

    def test_missing_optional_field(self):
        test_body = {
            "required_field": "Value 1"
        }
        valid, message = sanitize_fields(test_body)
        self.assertTrue(valid)

    def test_missing_required_field(self):
        test_body = {
            "optional_field": "Some value"
        }
        valid, message = sanitize_fields(test_body)
        self.assertFalse(valid)
        self.assertEqual("Field required_field is required", message)

    def test_extra_field(self):
        test_body = {
            "required_field": "Value 1",
            "extra_field": "Value 2"
        }
        valid, message = sanitize_fields(test_body)
        self.assertTrue(valid)
        self.assertTrue("extra_field" not in test_body, "extra_field removed")

    def test_field_too_long(self):
        test_body = {
            "required_field": "Value that exceeds 10 chars",
        }
        valid, message = sanitize_fields(test_body)
        self.assertFalse(valid)
        self.assertEqual("Field required_field exceeds maximum length of 10",
                         message)


class SanitizeFieldsHTTPErrorsTest(TestCase):
    def setUp(self):
        httpretty.enable()

    def tearDown(self):
        httpretty.disable()
        httpretty.reset()

    def test_404(self):
        httpretty.register_uri(httpretty.GET, 'http://example.com', status=404)
        with self.settings(REGS_GOV_API_URL='http://example.com'):
            valid, message = sanitize_fields({'something': 'else'})
        self.assertTrue(valid)

    def test_503(self):
        httpretty.register_uri(httpretty.GET, 'http://example.com', status=503)
        with self.settings(REGS_GOV_API_URL='http://example.com'):
            valid, message = sanitize_fields({'something': 'else'})
        self.assertTrue(valid)

    @patch('regulations.docket.requests.get')
    def test_timeout(self, get):
        get.side_effect = requests.Timeout
        with self.settings(REGS_GOV_API_URL='http://example.com'):
            valid, message = sanitize_fields({'something': 'else'})
        self.assertTrue(valid)



import re
from unittest import TestCase

from mock import patch
from django.test.client import Client

from regulations.generator import node_types


class ViewsSideBarViewTest(TestCase):
    """Integration tests for the sidebar"""

    @patch('regulations.views.sidebar.api_reader')
    def test_get(self, api_reader):
        api_reader.ApiReader.return_value.layer.return_value = {
            '876-12': [
                {'reference': ['1992-1', '876-12']},
                {'reference': ['1992-2', '876-12']},
            ],
            '876-12-a': [{'reference': ['1992-1', '876-12-a']}]
        }
        api_reader.ApiReader.return_value.regulation.return_value = {
            'label': ['876', '12'],
            'node_type': node_types.REGTEXT,
            'children': [
                {'label': ['876', '12', 'a'], 'children': [],
                 'node_type': node_types.REGTEXT},
                {'label': ['876', '12', 'b'], 'children': [],
                 'node_type': node_types.REGTEXT},
                {'label': ['876', '12', 'c'], 'node_type': node_types.REGTEXT,
                 'children': [
                     {'label': ['876', '12', 'c', '1'], 'children': [],
                      'node_type': node_types.REGTEXT}]}
            ]
        }
        response = Client().get('/partial/sidebar/1111-7/verver')

        sxs_start = response.content.find(b'<section id="sxs-list"')
        sxs_end = response.content.find(b'</section>', sxs_start)
        sxs = response.content[sxs_start:sxs_end]

        self.assertTrue(bool(re.search(br'\b12\b', sxs)))
        self.assertIn(b'12(a)', sxs)
        self.assertIn(b'1992-2', sxs)
        self.assertIn(b'876-12', sxs)
        self.assertIn(b'1992-1', sxs)
        self.assertIn(b'876-12-a', sxs)
        self.assertNotIn(b'876-12-b', sxs)



from datetime import datetime, timedelta
from unittest import TestCase

from mock import patch

from regulations.generator import versions


class VersionsTest(TestCase):
    @patch('regulations.generator.versions.api_reader')
    def test_fetch_grouped_history(self, reader):
        client = reader.ApiReader.return_value

        future = datetime.today() + timedelta(days=5)
        future = future.strftime('%Y-%m-%d')
        client.regversions.return_value = {'versions': [
            {'by_date': '2001-01-01', 'version': 'v1'},
            {'versions': 'v2'},
            {'by_date': '2003-03-03', 'version': 'v3'},
            {'versions': 'v4'},
            {'versions': 'v5'},
            {'by_date': future, 'version': 'v6'},
        ]}

        n1 = {'effective_on': '2001-01-01', 'publication_date': '2000-10-10'}
        n2 = {'effective_on': '2003-03-03', 'publication_date': '2001-01-01'}
        n3 = {'effective_on': '2003-03-03', 'publication_date': '2002-02-02'}
        n4 = {'effective_on': future, 'publication_date': '2005-05-05'}
        n5 = {'effective_on': future, 'publication_date': '2005-06-05'}
        n6 = {'effective_on': future, 'publication_date': '2005-07-05'}
        n7 = {'publication_date': '2001-02-02'}     # proposal
        client.notices.return_value = {'results': [n1, n2, n3, n5, n4, n6, n7]}

        history = versions.fetch_grouped_history('111')
        self.assertEqual(3, len(history))
        v1, v2, v3 = history

        self.assertEqual('future', v1['timeline'])
        self.assertEqual('current', v2['timeline'])
        self.assertEqual('past', v3['timeline'])

        self.assertEqual(versions.convert_to_python([n6, n5, n4]),
                         v1['notices'])
        self.assertEqual(versions.convert_to_python([n3, n2]), v2['notices'])
        self.assertEqual(versions.convert_to_python([n1]), v3['notices'])



from datetime import datetime
from unittest import TestCase

from regulations.generator.layers import utils


class LayerUtilsTest(TestCase):
    def test_convert_to_python(self):
        self.assertEqual("example", utils.convert_to_python("example"))
        self.assertEqual(1, utils.convert_to_python(1))
        self.assertEqual((1, 2.0, 8), utils.convert_to_python((1, 2.0, 8)))
        self.assertEqual(datetime(2001, 10, 11),
                         utils.convert_to_python('2001-10-11'))
        self.assertEqual(["test", "20020304", datetime(2008, 7, 20)],
                         utils.convert_to_python(['test', '20020304',
                                                  '2008-07-20']))
        self.assertEqual({'some': 3, 'then': datetime(1999, 10, 21)},
                         utils.convert_to_python({'some': 3,
                                                  'then': '1999-10-21'}))

    def test_is_contained_in(self):
        self.assertTrue(utils.is_contained_in('22-51-12', '22-51'))
        self.assertTrue(utils.is_contained_in('22-51-13', '22-51'))
        self.assertTrue(utils.is_contained_in('22-51', '22-51'))
        self.assertFalse(utils.is_contained_in('22-512', '22-51'))



from datetime import date
from unittest import TestCase

from django.test.client import Client
from mock import patch

from regulations.views.partial_search import PartialSearch


class PartialSearchTest(TestCase):
    """Integration tests for search"""

    @patch('regulations.views.partial_search.api_reader')
    @patch('regulations.views.partial_search.fetch_grouped_history')
    def test_get(self, fetch_grouped_history, api_reader):
        api_reader.ApiReader.return_value.search.return_value = {
            'total_hits': 3333,
            'results': [
                {'label': ['111', '22'], 'text': 'tttt', 'version': 'vvv',
                 'title':"consumer's"},
                {'label': ['111', '24', 'a'], 'text': 'o', 'version': 'vvv'},
                {'label': ['111', '25'], 'text': 'more', 'version': 'vvv'}
            ]
        }
        fetch_grouped_history.return_value = [
            {'notices': [{'document_number': 'bbb',
                          'effective_on': date(2012, 12, 12)}]},
            {'notices': [{'document_number': 'ccc',
                          'effective_on': date(2001, 1, 1)},
                         {'document_number': 'vvv',
                          'effective_on': date(2003, 4, 5)}],
             'timeline': 'timeytimey'}
        ]
        response = Client().get('/partial/search/111?version=vvv&q=none')
        self.assertIn(b'111-22', response.content)
        self.assertIn(b'111.22', response.content)
        self.assertIn(b'111-24-a', response.content)
        self.assertIn(b'111.24(a)', response.content)
        self.assertIn(b'111-25', response.content)
        self.assertIn(b'111.25', response.content)
        self.assertIn(b'3333', response.content)
        self.assertIn(b'Consumer&#39;s', response.content)

    @patch('regulations.views.partial_search.api_reader')
    @patch('regulations.views.partial_search.fetch_grouped_history')
    def test_root_info(self, fetch_grouped_history, api_reader):
        api_reader.ApiReader.return_value.search.return_value = {
            'total_hits': 2,
            'results': [
                {'label': ['444', '22'], 'text': 'tttt', 'version': 'vvv',
                 'title':"consumer's"},
                {'label': ['444', '24', 'a'], 'text': 'o', 'version': 'vvv'},
            ]
        }
        fetch_grouped_history.return_value = [
            {'notices': [{'document_number': 'bbb',
                          'effective_on': date(2012, 12, 12)}]},
            {'notices': [{'document_number': 'ccc',
                          'effective_on': date(2001, 1, 1)},
                         {'document_number': 'vvv',
                          'effective_on': date(2003, 4, 5)}],
             'timeline': 'timeytimey'}
        ]
        response = Client().get('/partial/search/444?version=vvv&q=none')
        self.assertIn(b'2 results', response.content)

    @patch('regulations.views.partial_search.api_reader')
    @patch('regulations.views.partial_search.fetch_grouped_history')
    @patch('regulations.generator.section_url.fetch_toc')
    def test_subinterp(self, fetch_toc, fetch_grouped_history, api_reader):
        fetch_toc.return_value = [
            {'index': ['444', 'Subpart', 'B'], 'is_subpart': True,
             'section_id': '444-Subpart-B', 'sub_toc': [
                {'index': ['444', '22'], 'section_id': '444-22',
                 'is_section': True}]},
            {'index': ['444', 'Interp'], 'section_id': '444-Interp',
             'sub_toc': [
                {'index': ['444', 'Interp', 'h1'],
                 'section_id': '444-Interp-h1'},
                {'index': ['444', 'Subpart', 'B', 'Interp'],
                 'section_id': '444-Subpart-B-Interp'}]}
        ]
        api_reader.ApiReader.return_value.search.return_value = {
            'total_hits': 3,
            'results': [
                {'label': ['444', '22', 'Interp'], 'text': 'tttt',
                 'version': 'vvv', 'title':"consumer's"},
                {'label': ['444', 'Interp', 'h1', 'p5'], 'text': 'o',
                 'version': 'vvv'}
            ]
        }
        fetch_grouped_history.return_value = [
            {'notices': [{'document_number': 'bbb',
                          'effective_on': date(2012, 12, 12)}]},
            {'notices': [{'document_number': 'ccc',
                          'effective_on': date(2001, 1, 1)},
                         {'document_number': 'vvv',
                          'effective_on': date(2003, 4, 5)}],
             'timeline': 'timeytimey'}
        ]

        response = Client().get('/partial/search/444?version=vvv&q=other')
        self.assertIn(b'444-Subpart-B-Interp', response.content)
        self.assertIn(b'444-Interp-h1', response.content)

    @patch('regulations.views.partial_search.api_reader')
    @patch('regulations.views.partial_search.fetch_grouped_history')
    def test_no_results(self, fetch_grouped_history, api_reader):
        api_reader.ApiReader.return_value.search.return_value = {
            'total_hits': 0,
            'results': []
        }
        fetch_grouped_history.return_value = [
            {'notices': [{'document_number': 'bbb',
                          'effective_on': date(2012, 12, 12)}]},
            {'notices': [{'document_number': 'ccc',
                          'effective_on': date(2001, 1, 1)},
                         {'document_number': 'vvv',
                          'effective_on': date(2003, 4, 5)}],
             'timeline': 'timeytimey'}
        ]
        response = Client().get('/partial/search/121?version=vvv&q=none')
        self.assertIn(b'4/5/2003', response.content)

    @patch('regulations.views.partial_search.fetch_grouped_history')
    def test_null_params(self, fetch_grouped_history):
        response = Client().get('/partial/search/111?version=vvv')
        self.assertIn(b'provide a query', response.content)
        response = Client().get('/partial/search/111?q=vvv')
        self.assertIn(b'provide a version', response.content)

    @patch('regulations.views.partial_search.api_reader')
    def test_preamble_search(self, api_reader):
        api_reader.ApiReader.return_value.search.return_value = {
            'total_hits': 3333,
            'results': [
                {'label': ['111_22', 'I', 'A'], 'text': 'tttt',
                 'title': 'A. Something'},
                {'label': ['111_22', 'I', 'p1'], 'text': 'eee'}
            ]
        }
        response = Client().get('/partial/search/preamble/111?q=none')
        self.assertIn(b'111_22-I-A', response.content)
        self.assertIn(b'111_22-I-p1', response.content)
        self.assertIn(b'tttt', response.content)
        self.assertIn(b'eee', response.content)
        self.assertIn(b'A. Something', response.content)
        self.assertIn(b'Section I.A', response.content)

    def test_add_prev_next(self):
        view = PartialSearch()
        context = {'results': {'total_hits': 77}}
        view.add_prev_next(0, context)
        self.assertFalse('previous' in context)
        self.assertEqual(context['next'], {'page': 1, 'length': 10})

        del context['next']
        view.add_prev_next(5, context)
        self.assertEqual(context['previous'], {'page': 4, 'length': 10})
        self.assertEqual(context['next'], {'page': 6, 'length': 10})

        del context['previous']
        del context['next']
        view.add_prev_next(6, context)
        self.assertEqual(context['previous'], {'page': 5, 'length': 10})
        self.assertEqual(context['next'], {'page': 7, 'length': 7})

        del context['previous']
        del context['next']
        view.add_prev_next(7, context)
        self.assertEqual(context['previous'], {'page': 6, 'length': 10})
        self.assertFalse('next' in context)



from django.template import loader

from regulations.generator.layers.base import InlineLayer
from regulations.generator.section_url import SectionUrl
from regulations.generator.layers import utils
from ..node_types import to_markup_id


class DefinitionsLayer(InlineLayer):
    shorthand = 'terms'
    data_source = 'terms'

    def __init__(self, layer):
        self.layer = layer
        self.template = loader.get_template(
            'regulations/layers/definition_citation.html')
        self.sectional = False
        self.version = None
        self.rev_urls = SectionUrl()
        self.rendered = {}
        # precomputation
        for def_struct in self.layer['referenced'].values():
            def_struct['reference_split'] = def_struct['reference'].split('-')

    def replacement_for(self, original, data):
        """ Create the link that takes you to the definition of the term. """
        citation = data['ref']
        # term = term w/o pluralization
        term = self.layer['referenced'][citation]['term']
        citation = self.layer['referenced'][citation]['reference_split']
        key = (original, tuple(citation))
        if key not in self.rendered:
            context = {'citation': {
                'url': self.rev_urls.fetch(citation, self.version,
                                           self.sectional),
                'label': original,
                'term': term,
                'definition_reference': '-'.join(to_markup_id(citation))}}
            rendered = utils.render_template(self.template, context)
            self.rendered[key] = rendered
        return self.rendered[key]



from django.db import models


class FailedCommentSubmission(models.Model):
    """ Model that holds comment submissions that failed when submitted to
    regulations.gov.
    """
    body = models.TextField()



import os
from selenium import webdriver
from testconfig import config


remote_configs = {
    'chrome': {
        'driver': webdriver.DesiredCapabilities.CHROME,
        'platform': 'LINUX',
        'version': '',
    },
    'ie11': {
        'driver': webdriver.DesiredCapabilities.INTERNETEXPLORER,
        'platform': 'Windows 10',
        'version': '11',
    }
}


class BaseTest():

    job_name = 'eRegs UI Test'

    def setUp(self):
        self.test_url = config['testUrl']
        self.driver = (
            self.make_remote()
            if 'remote' in config
            else self.make_local()
        )
        self.driver.set_window_size(800, 600)
        self.driver.implicitly_wait(30)

    def make_local(self):
        attr = config.get('local', 'PhantomJS')
        klass = getattr(webdriver, attr)
        if not isinstance(klass, type):
            raise TypeError(
                'Option {} did not resolve to a class'.format(attr))
        return klass()

    def make_remote(self):
        selenium_config = remote_configs[config['remote']]
        capabilities = selenium_config['driver']
        if (os.environ.get('TRAVIS')
                and os.environ.get('TRAVIS_SECURE_ENV_VARS')):
            capabilities.update({
                'tunnel-identifier': os.environ['TRAVIS_JOB_NUMBER'],
                'build': os.environ['TRAVIS_BUILD_NUMBER'],
            })

        username = os.environ['SAUCE_USERNAME']
        key = os.environ['SAUCE_ACCESS_KEY']
        capabilities['name'] = self.job_name
        capabilities['platform'] = selenium_config['platform']
        capabilities['version'] = selenium_config['version']
        hub_url = "%s:%s" % (username, key)
        executor = "http://%s@ondemand.saucelabs.com:80/wd/hub" % hub_url
        driver = webdriver.Remote(desired_capabilities=capabilities,
                                  command_executor=executor)
        jobid = driver.session_id
        print("Sauce Labs job: https://saucelabs.com/jobs/%s" % jobid)
        return driver

    def tearDown(self):
        self.driver.quit()



import tempfile
import os
import shutil

from regulations.generator.api_client import ApiClient
from unittest import TestCase


class ClientTest(TestCase):
    def test_local_filesystem(self):
        """ Verify that it's possible to host the files locally, where
        index.html is used in place of just the directory name. """
        tmp_root = tempfile.mkdtemp() + os.sep
        notice_path = tmp_root + os.sep + "notice" + os.sep
        os.mkdir(notice_path)
        with open(notice_path + "index.html", 'w') as f:
            f.write('{"results": ["example"]}')
        client = ApiClient()
        client.base_url = tmp_root
        results = client.get('notice')
        shutil.rmtree(tmp_root)
        self.assertEqual(["example"], results['results'])



from django.template import TemplateDoesNotExist
from django.template.loader import get_template

from regulations.generator.sidebar.base import SidebarBase
# @todo - seems like code in `generator` shouldn't reach in to `views`?
from regulations.views.utils import layer_names


class Help(SidebarBase):
    """Help info; composed of subtemplates defined by the active layers"""
    shorthand = 'help'

    def context(self, http_client, request):
        subtemplates = []
        for layer_name in sorted(layer_names(request)):
            template_name = 'regulations/sidebar/help/{}.html'.format(
                layer_name)
            try:
                template = get_template(template_name)
                subtemplates.append(template.render(
                    {'cfr_part': self.cfr_part}))
            except TemplateDoesNotExist:
                pass
        return {'subtemplates': subtemplates}



import unittest

from selenium.webdriver.support.ui import WebDriverWait

from regulations.uitests.base_test import BaseTest


class TOCTest(BaseTest, unittest.TestCase):

    job_name = 'TOC test'

    def test_toc(self):
        self.driver.set_window_size(1024, 600)
        self.driver.get(self.test_url + '/1005')
        html = self.driver.find_element_by_tag_name('html')
        WebDriverWait(self.driver, 30).until(
            lambda driver: 'selenium-start' in html.get_attribute('class'))

        drawer_toggle = self.driver.find_element_by_id('panel-link')
        drawer_toggle.click()

        # toggle arrow should switch
        self.assertTrue(drawer_toggle.get_attribute('class').find('open'))

        toc_link_1005_1 = self.driver.find_element_by_xpath(
            '//*[@id="toc"]/ol/li[1]/a')
        # toc link should have the proper section id attr
        self.assertEqual(toc_link_1005_1.get_attribute('data-section-id'),
                         '1005-1')
        toc_link_1005_1.click()

        # reg section should load in content area
        self.assertTrue(
            'catharine and myriads'
            in self.driver.find_element_by_class_name('section-title').text)

        # toc link should be highlighted
        self.assertTrue('current' in toc_link_1005_1.get_attribute('class'))

        WebDriverWait(self.driver, 90)

        # test another section
        toc_link_1005_3 = self.driver.find_element_by_xpath(
            '//*[@id="toc"]/ol/li[3]/a')
        self.assertEquals(toc_link_1005_3.get_attribute('data-section-id'),
                          '1005-3')

        toc_link_1005_3.click()
        # toc link should be highlighted
        self.assertIn('current', toc_link_1005_3.get_attribute('class'))

        WebDriverWait(self.driver, 10).until(
            lambda driver: 'clicked' in self.driver.find_element_by_class_name(
                'section-title').text)
        self.assertIn('current', toc_link_1005_3.get_attribute('class'))

        # make sure that the current class has been removed from the prev
        # section
        self.assertNotIn('current', toc_link_1005_1.get_attribute('class'))
